#ifndef _WORLDMODELCORE_H_
#define _WORLDMODELCORE_H_

#include "../serverInterface/MessageCoach.h"
#include "../../worldmodel/Field.h"
#include "../../worldmodel/Situation.h"
#include "../../basic/Ath98String.h"
class Sensors;

/**
  Kapselung des kompletten Wissen des Coaches.
 
  {\small {\bf Quelltext}

  [\URL[Header]{src/WorldModelCore.h} | [\URL[Implementierung]{src/WorldModelCore.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: WorldModelCore.h,v $
  *   $Revision: 1.6 $
  *   $Date: 2001/07/19 16:09:07 $
  *   $Author: wendler $ 
  *
  **************************************************
  \end{verbatim} }*/

class WorldModelCore
{
public:

  /** @name WorldModel anlegen und zerst\"oren */
  //@{  
    /** Erzeugt ein neues Weltmodell, das seine Informationen aus der
        Sensorik-Komponente Sensors bezieht.
        @param Pointer auf das aktive Sensorik-Objekt */
    WorldModelCore( Sensors *sensors );

    ~WorldModelCore();
  //@}


  /** @name Anfragen zum Agenten */
  //@{
    /** Liefert die Spielernummer des Agenten. */
    int playerNumber() const;
    
    /** Liefert den Namen des Teams, zu dem der Agent gehoert. */
    const Ath98String& teamName() const;
    
    /** Liefert die Seite des Agenten auf der das Tor seiner Mannschaft steht.
     */
    int side() const;

    /** Liefert die Id des Agenten. */
    int id() const;
  //@}


  /** @name Werte des Agenten setzen */
  //@{
    /** Setzt die Spielernummer des Agenten. */
    void setPlayerNumber( int newPlayerNumber, bool isGoalie );

    /** Setzt den Namen des Teams, zu dem der Agent gehoert. */
    void setTeamName( const Ath98String& newTeamName );

    /** Setzt die Seite, auf der der Agent spielt. */
    void setSide( int newSide );
  //@}


  /**@name Anfragen zur aktuellen und aelteren Situationen */
  //@{
    /** Liefert aktuelle und vergangene Situationen. \\
        Die haeufigste Verwendung ist sicher die Frage nach der aktuellen
        Situation. Der Parameter situationOffset ist per Default auf 0 gesetzt,
        so dass mit situation() auf die aktuelle Situation zugegriffen
        werden kann. \\
        @param situationOffset Mit situationOffset < 0 koennen vergangene
        (d.h. gespeicherte) Situationen abgerufen werden. Mit diesem offset
        wird die reale im SoccerServer vergangene Zeit angegeben, nicht der
        simstep-Wert.
        @exception Exception, wenn der
        situationOffset einen ungueltigen Wert hatte.
        @return Eine Referenz, die die angefragte Situation enth\"alt. Diese
        veraendert sich moeglicherweise mit Aenderung des simSteps(). */
    const Situation& situation( int situationOffset = 0 ) const;

    /** Liefert aktuelle Daten ueber den gewuenschten Spieler. 
        @param playerId die Id des gewuenschten Spielers.
        @exception Situation::NoPlayerWithThisIdException, wenn der angegebene
        Spieler in der aktuellen Situation nicht existiert.
        @return ModelledPlayer mit den gewuenschten Daten. */
    const ModelledPlayer& player( int playerId ) const;

    /** Liefert aktuelle Daten ueber den Ball. */
    const ModelledBall& ball() const;

    /** Liefert den Simulationstakt der aktuellen Situation. 
        Die zu diesem Zeittakt gehoerende Situation wird von situation() 
        geliefert. */
    int simStep() const;

    /** Liefert den Spielzustand zurueck. Dieser ist als int kodiert. */
    int gameState() const;

    /** Liefert die Initiative.
        @return Einer der Werte OUR_TEAM_ID, THEIR_TEAM_ID, BOTH_TEAMS_ID,
        NO_TEAM_ID */
    int initiative() const;

    /** Gibt an, ob ein bestimmtes Team die Initiative hat. 
        Diese Funktion erleichtert das Umgehen mit der Initiative, da mehrfache
        Abfragen unn\"otig werden. Wir haben die Initiative, wenn, die
        Initiative auf OUR_TEAM_ID oder auf BOTH_TEAMS_ID steht.
        Entsprechendes gilt fuer THEIR_TEAM_ID.
        @param aTeamId Das Team, von erfragt werden soll, ob es die Initiative
        hat. M\"ogliche Werte sind OUR_TEAM_ID, THEIR_TEAM_ID.
        @exception InvalidIdException, wenn die aTeamId einen ungueltigen Wert 
        hat.
        @return YES oder NO. */
    bool hasInitiative( int aTeamId ) const;

    /** Anfrage ob wir uns gerade in einem Spielzustand befinden in dem
        die Zeit im SoccerServer in jedem Simulationstakt weitergezaehlt
        wird oder nicht
        @return YES wenn die Zeit im aktuellen Spielzustand fortlaeuft
        @return NO wenn die Zeit im aktuellen Spielzustand still steht */
    bool isTimeElapsingGameState() const;
  //@}

  /** @name Werte der aktuellen Situation setzen */
  //@{
    /** Setzt den Simulationstakt der aktuellen Situation. */
    void setSimStep( int newValue );

    /** Setzt den Spielzustand. */
    void setGameState( int newGameState );
    
    /** Setzt die Initiative. 
        Dabei kann auch eine Seite angegeben werden.  Spielen wir gerade auf
        dieser Seite, wird die Initiative auf OUR_TEAM_ID gesetzt, andernfalls
        auf THEIR_TEAM_ID. \\
        @param newValue Entweder die Id eines Teams ( OUR_TEAM_ID, 
        THEIR_TEAM_ID, BOTH_TEAMS_ID ) oder die Id einer Seite (LEFT_SIDE_ID, 
        RIGHT_SIDE_ID ) */
    void setInitiative( int newValue );
  //@}


  /**@name Allgemeine Allgemeine Anfragen und Setzmoeglichkeiten */
  //@{
    /** Liefert die Zahl der Tore, die ein Team bisher erzielt hat.
        @param aTeamId Id des Teams. Gueltig sind OUR_TEAM_ID und THEIR_TEAM_ID
        @exception InvalidIdException, wenn die aTeamId einen ungueltigen Wert 
        hat.
        @return Die Zahl der Tore dieses Teams. */
    int score(int aTeamId) const;

    /** Setzt die Zahl der Tore fuer ein Team.
        @param newScore Die neue Zahl der Tore
        @param aTeamId Id des Teams. Gueltig sind OUR_TEAM_ID und THEIR_TEAM_ID
        @exception InvalidIdException, wenn die aTeamId einen ungueltigen Wert 
        hat */
    void setScore( int newScore, int aTeamId );

    /** Liefert das Feld, auf dem der Agent spielt. */
    const Field& field() const;
  //@}


  /** @name Abfragen zur Vorzugsrichtung */
  //@{
    /** Liefert die bevorzugte Kickrichtung.
        Diese "Vorzugsrichtung" ist abh\"angig von der Position des Spielers.
        Durch sie werden wichtige Eigenschaften des Mannschaftsspiels
        umgesetzt, z.b. Flanken und Angriffe \"uber die Fl\"ugel.
        @param pos Position, f\"ur die die Vorzugsrichtung berechnet werden
        soll.
        @return Ein Richtungsvektor. Die Richtungskomponente ( \Ref{length()})
        direkt der kickBall-Skill \"ubergeben werden. Die Laenge gibt die
        optimale Ballgeschwindigkeit in m/simStep an. Die Division der Laenge
        durch globalConf->kickPowerRate ergibt eine Kraft, die kickBall
        \"ubergeben werden kann.*/
    Vector preferredKickDirection( const Vector& pos ) const;

    /** Liefert die bezorzugte R\"uckpassrichtung.
        @param pos Position, f\"ur die die Vorzugsrichtung berechnet werden
        soll.
        @return Ein Richtungsvektor. Die Richtungskomponente ( \Ref{length()})
        direkt der kickBall-Skill \"ubergeben werden. Die Laenge gibt die
        optimale Ballgeschwindigkeit in m/simStep an. Die Division der Laenge
        durch globalConf->kickPowerRate ergibt eine Kraft, die kickBall
        \"ubergeben werden kann.*/
    Vector backKickDirection( const Vector& pos ) const;

 private:
    /** \"Ausseres Limit. 
        y-Koordinate einer waagrechten Linie 6m vom Feldrand. */
    static const double OUTER_LIMIT;
    /** Inneres Limit. 
        y-Koordinate einer waagrechten Linie 14m vom Feldrand. */
    static const double INNER_LIMIT;

    /** @name RegionOnField
        Regionen auf dem Feld
        Symbolische Bezeichner f\"ur die Aufteilung des Feldes in verschiedene
        Regionen, in denen die preferredKickDirection verschieden berechnet
        wird.
        @see preferredKickDirection
    */
    enum RegionOnField { UPPER = 6, UPPER_MIDDLE = 4,
                         LOWER = 7, LOWER_MIDDLE = 5,
                         TO_UPPER = 1, TO_LOWER = 2,
                         REST = 3,
                         UPPER_EDGE = 9, LOWER_EDGE = 10,
                         ENEMY_GOAL = 8 };

    /** Liefert die Region, in der sich eine Position befindet.
        Ja nach Region kann sich ein Spieler anders verhalten. Auch die
        Definition der Vorzugsrichtungsfunktion basiert auf der Region, in der
        sich eine Position befindet. 
        @param pos Position, f\"ur die die Region ermittelt werden soll.
        @return Die Nummer der Region, ein Enum vom Type RegionOnField.
        @see RegionOnField
    */
    RegionOnField region( const Vector& pos ) const;
 public:
  //@}


  /** @name Einarbeitung und Kommunikation */
  //@{
    /** Integriert die aktuellen Sensorinformationen in das Weltmodell. */
    void integrateSensorInfo();

 private:
    void processRefereeInfo();
    void processRefereeMessage( const MessageCoach& message );
    void processHearInfo();
    void processChangePlayerTypeInfo();
 public:
    /** Erzeugt eine Nachricht fuer den Ruf des Weltmodells.
        @return die Nachricht */
    Ath98String saySituation();
  //@}


  /** @name Methoden zum Debugging */
  //@{
    /** Liefert einen String fuer die Klassenbezeichnung zurueck. */
    string idString() const;

    /** Ausgabe auf einen String-Stream */
    friend ostream& operator<< ( ostream&, const WorldModelCore& );
  //@}


private:
  /** @name Instanzvariablen */
  //@{
    /** Zeiger auf das Sensorik-Objekt */
    Sensors *theSensors;

    /** Das Feld, auf dem der Spieler spielt. */
    Field theField;

    /** Array, das vergangene (und zukuenftige) Situationen enthaelt. */
    vector<Situation> theSituations;

    /** Index der aktuellen Situation im Array */
    int theCurrentSituationIndex;

    /** Interne Feldgroesse fuer das Array der Situationen. */
    int theSituationFieldSize;

    /** Der Teamname */
    Ath98String theTeamName;

    /** Die Spielernummer */
    int thePlayerNumber;

    /** Die Seite des Spielfelds, auf der das Tor des Teams des Agenten 
        steht. */
    int theSide;

    /** Die Zahl der bisher gefallenen Tore  */
    int theScore[2];
    
    /** Das Team, das die Initiative besitzt. */
    int theInitiative;
    
    /** Der aktuelle Spielzustand */
    int theGameState;

    /** String der die Kurzkennung der Klasse enthaelt */
    string theIdString; 
  //@}
};

#endif
