#ifndef _SKILLS_H_
#define _SKILLS_H_

#include <math.h>
#include <string>
#include <vector>

#include "../../basic/Vector.h"
#include "../../worldmodel/ModelledBall.h"
#include "../../worldmodel/ModelledPlayer.h"
#include "../../worldmodel/WorldModelInterface.h"

/**
  Die F\"ahigkeiten des Agenten. 
  Options verwendet Skills, um Plaene zu erzeugen. Ein Plan ist dabei eine
  Folge von konkreten Aktionen, die an den SoccerServer geschickt werden 
  koennen.

  Im Skills-Objekt gibt es eine Ausgangssituation, die anf\"anglich auf die
  aktuelle Situation des Weltmodells gesetzt ist. Jeder Aufruf einer F\"ahigkeit
  f\"uhrt zu \"Anderungen an dieser Situation, so da\3 der beim Aufruf der
  n\"achsten F\"ahigkeit die neue, modifizierte Situation als Ausgangssituation
  genomment werden kann. Vor der Berechnung jedes Plans mu\3 die
  Ausgangssituation mit \Ref{resetSituation}() auf die Gegenwart zur\"uckgesetzt
  werden. Eine Referenz auf die Situation ist intern \"uber die private Methode
  \Ref{situation}() verf\"ugbar.
  
  {\small {\bf Quelltext}

  [\URL[Header]{src/Skills.h} | [\URL[Implementierung]{src/Skills.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: Skills.h,v $
  *   $Revision: 1.1.1.1 $
  *   $Date: 2001/07/31 22:23:09 $
  *   $Author: rami $ 
  *
  **************************************************
  \end{verbatim} }*/
class Skills
{
public:

  /** Minimale Drehung, die ausgef\"uhrt werden soll. */
  static const double MIN_TURN_ANGLE_IN_DEGREE = 1.0;

  /** Konstante zur Steuerung des Verhaltens f"ur gotoRegion.
    Die Konstanten k\"onnen addiert werden, um mehrere
    Verhaltensweisen gleichzeitig angeben zu k\"onnen. 
    @see gotoRegion */
  static const int STANDARD_BEHAVIOR = 0;
   /** Konstante zur Steuerung des Verhaltens f"ur gotoRegion.
    @see gotoRegion */
  static const int AVOID_OBSTACLES = 1;
  /** Konstante zur Steuerung des Verhaltens f"ur gotoRegion.
    @see gotoRegion */
  static const int RUN_BACKWARDS = AVOID_OBSTACLES<<1;
  /** Konstante zur Steuerung des Verhaltens f"ur gotoRegion.
    @see gotoRegion */
  static const int AVOID_BALL_AND_PLAYERS = RUN_BACKWARDS<<1;



  /** Initialisiert ein Skills-Objekt.
    Es ist nicht sinnvoll mehrere Skills-Objekte zu haben. */
  Skills();




  /** Berechnet den Plan, um zu einer gegebenen Region zu gelangen.
    Diese Region ist kreisfoermig und wird durch einen Mittelpunkt und einen
    Radius bestimmt. Der Radius darf nicht kleiner als TRIANGULATION_ACCURACY
    sein.
    @param centerOfTarget Mittelpunkt der Zielregion.
    @param targetRadius Der Radius der Zielregion. 
    @param staminaLimit die untere Staminagrenze (default 0)
    @param specialBehavior. Hier kann eine Summe von Verhaltenskonstanten
    angegeben werden. M\"oglich sind bisher: AVOID_OBSTACLES; wenn dieser Wert 
    angegeben ist, werden Hindernisse (also andere Spieler) umgangen (default).
    AVOID_BALL_AND_PLAYERS; ist eine Erweiterung von AVOID_OBSTACLES,
    wo auch der Ball als Hindernis angesehen wird.
    RUN_BACKWARDS; ist dieser Wert angegeben, rennt der Spieler r\"uckw\"arts
    (was wesentlich langsamer ist). STANDARD_BEHAVIOR: Halt nix besondres.
    @param desiredMaxLengthOfPlan gew\"unschte L\"ange des Planes. 
    (default UNDEF_INT was dann maxLengthOfPlan ergibt). */

  vector<Action> gotoRegion( const Vector& centerOfTarget,
      double targetRadius = TRIANGULATION_ACCURACY,
      const double staminaLimit = 0.0, 
      const int specialBehavior = AVOID_OBSTACLES,
      double desiredMaxPower = 100.0,
      int maxDesiredSteps = 50);

  /** Nur einen Kick machen aufs Ziel. */
  Action kickOnce( const Vector& aTarget );

  /** intelligenter CompoundKick */
  Action kick(const Vector& aTarget, double desiredTargetSpeed, int maxSteps);

  Action dribble( const Vector aTarget );
  


  /** Liefert die Aktion fuer das Rufen einer Nachricht. */
  Action simpleSay( const Ath98String& message );

  /** Berechnet die Action fuer den WorldModell-Ruf. */
  Action saySituationAction( const Situation& aSituation = globalWorldModel->situation() );

  /** Liefert eine leere Aktion. */
  Action simpleEmptyAction( );

  /** Liefert die Aktion fuer eine Beschleunigung um die angegebene Staerke. */
  Action simpleDash( const double power );

  /** Berechnet den Plan fuer eine Drehung um einen Winkel.
    @param turnAngle Der Winkel um den sich der Spieler drehen
    soll. Ist die Drehung in einem Takt nicht mglich, werden
    mehrere Aktionen zurckgeliefert, bis die gewnschte Drehung
    um turnAngle erfolgt ist.
    der Winkel liegt im 2001-Format vor -> 0 - 360
    @return Ein Plan, der aus mehreren Turn-Aktionen besteht.
    @see turnTo */
  vector<Action> turnBy( double turnAngle );

  /** Berechnet den Plan fuer eine Drehung auf einen Winkel.
    @param finalDirection Die gewuenschte neue Blickrichtung. Das ist der Winkel
    eines Vektors von der absoluten Spielerposition zu einer absoluten
    Zielposition. direction ist also unabh\"angig von der momentanen
    Blickrichtung des Spielers.
    der Winkel liegt im 2001-Format vor -> 0 - 360
    @return Ein Plan, der aus mehreren Turn-Aktionen besteht.
    @see turnBy */
  vector<Action> turnTo( const Vector target );

  /** Liefert den Plan fuer eine Drehung um den angegebenen Winkel. */
  Action simpleTurn( const double angle );

  /** Liefert die Aktion fuer das Greifen nach dem Ball (nur fuer Torwart). */
  Action simpleCatch( const double angle );
  
  /** Setzt die Ausgangssituation f\"ur die einzelnen Skills zur\"uck.
    Nach dem Aufruf dieser Routine ist die Ausgangssituation die aktuelle
    Situation des Weltmodells. 
    
    Diese Routine mu\3 unbedingt jedesmal aufgerufen werden, bevor ein 
    Plan erstellt wird, der von der Gegenwart ausgeht. Jeder Aufruf
    einer F\"ahigkeit \"andert dann die Ausgangssituation.
    @param scenario Szenario welches die Ausgangssituation auf ein anderes
    als das aktuelle Weltbildszenario setzt. */
  void resetSituation( Situation scenario = globalWorldModel->situation() );
  /** @name Kick-Tabellen. */

  /** Liefert einen String mit Klassenbezeichnung zurueck. */
  string idString() const;

  /** */
  double getNewBallDistance(int i);

  /** Liefert den ModelledPlayer des Agenten der Situation. */
  const ModelledPlayer& player() const;

  /** Liefert den ball der Situation. */
  const ModelledBall& ball() const;



protected :
    
  /** String der die Kurzkennung der Klasse enthlt */
  string theIdString;

private:
  /** Liefert die Situation, von der die n\"achste Skill-Routine ausgeht.
    Der simStep dieser Situation ist gr\"osser/gleich der aktuellen Situation
    im Weltmodell.
    @return eine Referenz auf die Ausgangssituation. Sie darf modifiziert 
    werden. */
  Situation& situation();


  /** Situation, auf der die einzelnen Skills arbeiten */
  Situation theSituation;


  /** Tabellen mit Ballgeschwindigkeiten und -entfernungen
      fuer Auskunftsfunktionen. Die Tabellen werden einmal
      beim Start des Agenten berechnet */
  
  /** Verzoegerungsfaktor fuer die Ballgeschwindigkeit
      ermittelt aus Versuchen mit dem Learnserver */
  static const double theBallVelocityFactor =  0.941843;
  
  /** Dimension der Tabellen theBallDistanceTable und 
      theBallVelocityTable*/
  static const int theBallTableMaxIndex = 60;
  
  /** x-Dimension der Tabelle theBallStartVelocityTable */
  static const int theStartVelocityTableMaxDistance = 43;
  
  /** y-Dimension der Tabelle theBallStartVelocityTable */
  static const int theStartVelocityTableMaxSteps = 60;
  
  /** Diese Tabelle enthaelt Ballentfernungen, als Index 
      dient der Simstep */
  double theBallDistanceTable[theBallTableMaxIndex];

  /** Diese Tabelle enthaelt Ballgeschwinidigkeiten, 
      als Index dient der Simstep */
  double theBallVelocityTable[theBallTableMaxIndex];

  /** Diese Tabelle enthaelt Ballstartgeschwinidigkeiten, 
      als Indizes dienen die gewuenschte Distanz und
      die Anzahl der zur Verfuegung stehenden Simsteps */
  double theBallStartVelocityTable[theStartVelocityTableMaxDistance]
    [theStartVelocityTableMaxSteps];

  /** Diese neue! Tabelle enthaelt Ballentfernungen, als Index dient der Simstep */
  double theNewBallDistanceTable[101];

  /** Diese neue! Tabelle enthaelt Ballentfernungen, als Index dient der Simstep */
  void createNewBallDistanceTable();

  /** schreibt die Tabelle auf stdout */
  inline void Skills::printNewBallDistanceTable() const;
  
  /** Diese Methode erstellt alle Tabellen */
  void createBallTables();

  /** Diese Methode schreibt alle Tabellen auf STDOUT */
  void printBallTables();

  double theRelativeBallDistanceTable[50];

  double theMaxPlayerDistance[50];

  };


/** Global verfuegbare Skills. */
extern Skills *globalSkills;


inline Situation& Skills::situation()
{
  return theSituation;
}

inline const ModelledPlayer& Skills::player() const
{
  return theSituation.playerWithId( globalWorldModel->id() );
}

inline const ModelledBall& Skills::ball() const
{
  return theSituation.ball();
}

inline string Skills::idString() const
{
  return theIdString;
}


inline void Skills::createNewBallDistanceTable() {
  double v = 1.0; 
  double s = 0.0;
  for( int i=0; i< 101; ++i){
    theNewBallDistanceTable[i] = s;
    s += v;
    v *= globalConf->ballDecay;
  }
}

inline void Skills::printNewBallDistanceTable() const {
  for( int i=0; i< 51; ++i){
    cout << i << '\t' << theNewBallDistanceTable[i] << endl;
  }
}

inline double Skills::getNewBallDistance(int i) {
  if( (i < 0) || (i > 100) ){
    cerr << globalWorldModel->id() << " - Skills::getNewBallDistance  i out of range --> returning: theNewBallDistanceTable[100]  !!!\n";
    i = 100;
  }
  return theNewBallDistanceTable[i];
}

#endif
