#ifndef _TARGET_H_
#define _TARGET_H_

#include "../../basic/Configuration.h"
#include "../../worldmodel/WorldModelInterface.h"

#include <iostream>

/**
   Die Klasse Target dient zur Beschreibung eines Positions-Ziels, das der
   Agent anstrebt.
   Diese Ziele sind durch Orte oder Objekte auf dem Spielfeld definiert. D.h.
   neben festen Positionen kann ein Ziel auch beweglich sein (z.B. der Ball).
   Zur Repr\"asentation eines Ziels ist es notwendig, da\3 die konkrete
   Position jeweils aktuell bestimmt werden kann. Aus diesem Grund kann ein
   Target sowohl mit einer festen Positionsangabe als auch durch die Angabe
   einer Target-Id konstruiert werden.
   Wird eine feste Position \"ubergeben, so wird die Target-Id automatisch
   auf den Wert "NOTHING" gesetzt. Ansonsten wird die Id abgespeichert.
   Soll nun auf die Position des Target zugegriffen werden, so wird beim
   Zugriff die Methode \Ref{pos} aufgerufen, die, der Target-Id entsprechend,
   die aktuelle Position des Objektes liefert. Bei einer festen Position
   wird diese zur\"uckgegeben.
   Die Realisierung der Berechnungen der konkreten Position eines Objektes
   erfolgt in der ihm zugeh\"origen "calculate..."-Methode und sollte nach
   M\"oglichkeit keine zeitkritischen Berechnungen enthalten.
   \"Uber die einfache Berechnung eines Zielpunktes k\"onnen durch die
   Definition eines Targets jederzeit auch dynamische Elemente der
   Spielsituation einflie\3en.

  {\small {\bf Quelltext}

  [\URL[Header]{src/Target.h} | [\URL[Implementierung]{src/Target.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: Target.h,v $
  *   $Revision: 1.1.1.1 $
  *   $Date: 2001/07/31 22:23:09 $
  *   $Author: rami $ 
  *
  **************************************************
  \end{verbatim} }*/


class Target{
 public:
  typedef int TargetId;
  /** @name Variablen, die die Art des Ziels beschreiben. */

  //@{
  ///
  static const TargetId NOTHING = 0;
  ///
  static const TargetId BALL = 1;
  ///
  static const TargetId INTERCEPT = BALL<<1;
  ///
  static const TargetId INTERCEPT_OFFENSIV = INTERCEPT<<1;
  ///
  static const TargetId GOALIE_EMERGENCY_KICK = INTERCEPT_OFFENSIV<<1;
  ///
  static const TargetId GOAL_KICK_AFTER_CATCH_PASS_TARGET = GOALIE_EMERGENCY_KICK<<1;
  ///
  static const TargetId KICK_FREE_KICK = GOAL_KICK_AFTER_CATCH_PASS_TARGET<<1;
  ///
  static const TargetId PASS = KICK_FREE_KICK<<1;
  ///
  static const TargetId DP_PASS = PASS<<1;
  ///
  static const TargetId PLAYER = DP_PASS<<1;
  ///
  static const TargetId SCORE_GOAL = PLAYER<<1;
  ///
  static const TargetId POSITION = SCORE_GOAL<<1;
  ///
  static const TargetId GOALIE_STANDARD_POSITION = POSITION<<1;
  ///
  static const TargetId GOALIE_INTERCEPT_POSITION = GOALIE_STANDARD_POSITION<<1;
  ///
  static const TargetId GOALIE_KICK_TO_THE_OTHER_SIDE = GOALIE_INTERCEPT_POSITION<<1;
  ///
  static const TargetId GOALIE_MOVE = GOALIE_KICK_TO_THE_OTHER_SIDE<<1;
  ///
  static const TargetId OUR_FREEKICK = GOALIE_MOVE<<1;
  ///
  static const TargetId THEIR_FREEKICK = OUR_FREEKICK<<1;
  ///
  static const TargetId DRIBBLE = THEIR_FREEKICK<<1;
  ///
  static const TargetId COVER = DRIBBLE<<1;
  ///
  static const TargetId CHAIN = COVER<<1;
  //@}

  /** die Intercept-modes */
  static const short FASTEST = 1;
  static const short SHORTEST = 2;
  static const short OFFENSIVE = 4;
  

  /*@name Konstruktoren */
  //@{
  /** Default-Konstruktor. */
  Target();
  /** Konstruktor f\"ur ein festes Ziel.
      @param aPos eine Zielposition auf dem Feld
      @param aTime die Zeit in Simsteps, die das Ziel glig sein soll
      Default: UNDEF_INT (wird z.Zt. nicht benutzt). */
  Target( const Vector& aPos, int aTime = UNDEF_INT );

  /** Konstruktor fr ein variables (i.d.R. bewegliches) Ziel. Es wird eine
      Id vom Typ TargetId \"ubergeben, die das Ziel beschreibt.
      @param aId die Target-Id
      @param aTime die Zeit in Simsteps, die das Ziel glig sein soll
      Default: UNDEF_INT (wird z.Zt. nicht benutzt). */
  Target( TargetId aId );

  /** Konstruktor fr ein Spielerziel.
      @param aId die Target-Id
      @param aPlayerId die Id des Spielers, der das Zeil ist
      @param aTime die Zeit in Simsteps, die das Ziel glig sein soll
      Default: UNDEF_INT (wird z.Zt. nicht benutzt). */      
  Target( TargetId aId, int aPlayerId, int aTime = UNDEF_INT );

  //@}
  /**@name Zugriffsmethoden */
  //@{
  /** Abrage der aktuellen Position des Targets. Anhand der Target-Id wird
      die aktuelle Position des Targets bestimmt. Dazu wird die zugeh\"orige
      Berechnungsmethode aufgerufen. */
  Vector pos() const;

  /** Abfrage des Zeitstempels des Targets. */
    int time() const;

  /** Liefert die Id des targets zur\"uck. Diese kann u.a. dazu benutzt
      werden, um innerhalb einer untergeordneten Intention festzustellen,
      welcher Art das angestebte Ziel ist, um auf dieser Basis eine
      besondere Verhaltensvariante zu generieren. */
  const TargetId id() const;
  //@}
  /**@name Zuweisungsmethoden */
  //@{
  /** Setzen einer festen Spielfeldposition. Die Target-Id wird dabei auf
      "NOTHING" gesetzt. */
  void setPos( const Vector& aPos );

  /** Setzen eines Zeitstempels. */
  void setTime( int aTime );
  //@}

  /**@name Sonstige Methoden der \"offentlichen Schnittstelle */
  //@{
  /** Vergleichsoperator, der das Target mit einem anderen Target
      vergleichen kann. Es wird gepr\"uft, ob die beiden Targets weiter
      als "globalConf->interceptBallAccuracy" voneinander entfernt sind. */
  bool isNearlyEqual( const Target& aTarget );
  //@}

 private:
  /**@name Variablen */
  //@{
  /** Die Position des Targets. */
  mutable Vector thePos;

  /** Die Target-Id. */
  TargetId theId;

  /** Die Spieler-Id. */
  int thePlayerId;

  /** Der Zeitstempel. */
  int theTime;

  /** Der InterceptMode. */
  short theInterceptMode;

  /** Die Takte fuer den OffensiveIntercept */
  int theNumberOfSimSteps;

  /** Zeit der letzten Berechnung. */
  mutable int theTimeOfLastCalc;


  //@}

  /**@name Methoden zur Berechnung von dynamischen Targets */
  //@{

  // fr die Ppositionierung
  Vector SBSPPosition( int playerNumber ) const;
  Vector Target::SBSPPosition2( int playerNumber ) const;

  static const int CP_at_point_buffer = 1;
  inline float XToOnsideX( float x, float buffer );
  inline float XToOnsideX( float x );
  inline Vector positionToOnsidePosition( Vector pos, float buffer );
  inline Vector positionToOnsidePosition( Vector pos );
  

  /** Berechnung des exacten Freekick-Targets. Diese Berechnung
      wird durchgef\"uhrt, wenn der Spieler bereits dicht am Ball ist.
      Berechnet wird eine Position, die 0.6 Meter von Ball entfernt
      Richtung prefferedKickDirection liegt. */
  Vector calculatePositionToKickFree() const;

  /** Berechnung fuer die Nahpositionierung beim eigenen Freisto\3. */
  void calculateOurFreekick() const;

  /** Berechnung fuer DP_Pass. */
  Vector calculateDPPass() const;

  /** Berechnung fuer Kettenbildung. */
  Vector calculateChain() const;

  /** Berechnung fuer Torschu. */
  Vector calcScoreGoal() const;

  /** Berechnung des Deckungspunktes. */
  Vector calcCover() const;

  int estimatedBallSteps( double dist ) const;
  //@}
};


inline float Target::XToOnsideX( float x, float buffer )
{
  return ( (float) min( (double) x, globalWorldModel->situation().offsideLineX() - buffer ) );
}

inline float Target::XToOnsideX( float x )
{
  return XToOnsideX( x, CP_at_point_buffer );
}

inline Vector Target::positionToOnsidePosition( Vector pos, float buffer )
{
  pos.setX( XToOnsideX( pos.x(), buffer ) );
  return pos;
}

inline Vector Target::positionToOnsidePosition( Vector pos )
{
  return positionToOnsidePosition( pos, CP_at_point_buffer );
}  


inline ostream& operator<<( ostream& os, const Target& aTarget ){
  os << aTarget.pos() << " with Id: " << aTarget.id() << '\n';
  return os;
}

inline void Target::setPos( const Vector& aPos ){
  thePos = aPos;
  theId = NOTHING;
}

inline int Target::time() const{
  return theTime;
}

inline void Target::setTime( int aTime ){
  theTime = aTime;
}

inline bool Target::isNearlyEqual( const Target& aTarget ){
  if( !thePos.isValid() ) return false;
  return ( thePos-aTarget.pos() ).length() <= globalConf->interceptBallAccuracy;
}

inline const Target::TargetId Target::id() const{
  return theId;
}

#endif
