#ifndef _MESSAGE_H_
#define _MESSAGE_H_

#include "../basic/Ath98String.h"
#include "../basic/Common.h"

/**
  Kapselung einer Nachricht die vom Schiedsrichter kommt oder zwischen
  Spielern ueber den SoccerServer ausgetauscht werden kann. 
  Eine Nachricht besteht aus der eigentlichen Mitteilung (einem Ath98String), 
  einer Zeit fuer die sie gilt und einem Absender. 
  Der Absender ist entweder eine Spieler- oder die Schiedsrichter-ID. 
  
  {\small {\bf Quelltext}

  [\URL[Header]{src/Message.h} | [\URL[Implementierung]{src/Message.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: Message.h,v $
  *   $Revision: 1.1.1.1 $
  *   $Date: 2001/07/31 22:23:09 $
  *   $Author: rami $
  *
  **************************************************
  \end{verbatim} }*/
class Message
{

  friend class MessageTest;

public:

  /** Aufbau einer gehoerten Nachricht die vom SoccerServer gekommen ist.
      Initialisierung mit einem Ath98String. */
  Message( const Ath98String& information );

  /** Generierung einer Nachricht, die an den SoccerServer
      gesendet werden soll. */
  Message( const Ath98String& information, long simStep );

  /** Konstrukter fuer Template-Klassen */
  Message();

  virtual ~Message();

  /** Setzt den Teamname fuer die Verschluesselung der Nachrichten. */
  static void setTeamname( const Ath98String& teamname );

  /** Aufbau einer gehoerten Nachricht die vom SoccerServer gekommen ist.
      Initialisierung mit einem Ath98String. */
  void setByHearInfo( const Ath98String& information );

  /** Generierung einer Nachricht, die an den SoccerServer
      gesendet werden soll. */
  void setBySayInfo( const Ath98String& information, long simStep );

  /** Die Zeit zu der die Nachricht gehoert.
      Bei einer empfangenen Nachricht vom Schiedsrichter ist dies die
      Zeit des Empfangs der Nachricht. Bei einer empfangenen Nachricht
      eines Spielers ist dies die Zeit die zur empfangenen Nachricht
      gehoert. Beim Absenden ist dies die Zeit zu der die zu sendene
      Nachricht gehoert. */
  long time() const;

  /** Liefert die Nachricht unverschluesselt zurueck. */
  const Ath98String& message() const;

  /** Liefert Infos ueber den Rufer. Moegliche Rueckgabewerte sind:  SELF_ID,
    OUR_TEAM_ID (eigenes Team, aber nicht der Agent selbst),
    THEIR_TEAM_ID, REFEREE_ID, UNDEF_TEAM_ID (im Fehlerfall). */
  int senderInfo() const;

  /** Liefert eine zum Senden aufbereitete Version der Nachricht. */
  Ath98String messageForSend() const;

  /** Ausgabe auf einen String-Stream */
  friend ostream& operator<< (ostream&, const Message& );

protected:

  /** Setzt alle Daten, die man dem Ath98String entnehmen kann. */
  virtual void parseHearMessage(const Ath98String& information);

  /** Verschluesselt die Nachricht zum Senden an den SoccerServer
      und gibt die verschluesselte Nachricht zurueck. 
      Eine solche Nachricht kann maximal eine Laenge von 378 Byte haben.
      (378 (message) + 6 (header) = 384 -> 512 Byte in Server-Format.
      Nachrichten mit groessere Laenge werden abgeschnitten und erzeugt. */
  Ath98String encodeSayMessage( const Ath98String& information, 
                                long simStep ) const;

  /** Erzeugt eine Nachricht aus einem empfangenen Ruf.
      Wenn die Nachricht message nicht vom eigenen Team stammte 
      enthaelt die mit message() erfragbare Information einen
      leeren Ath98String, ansonsten steht dort die Nachricht. */
  void decodeHearMessage( const Ath98String& information );

  /** Verschl\"usselt die naechsten 3 8chars von in zu 4 6chars von out */
  void encode( const unsigned char* in, unsigned char*& out ) const; 
  
  /** Entschl\"usselt die naechsten 4 6chars von in zu 3 8chars von out*/
  void decode( const unsigned char* in, unsigned char*& out ) const; 

  /** Generiert eine Pruefsumme fuer in (gegebenen in 8chars) und 
   haengt an out 3 8chars ran. */
  void addCheckSum( const unsigned char* in, const int lengthOfIn, 
                    unsigned char* out ) const;  

  /** Haengt den aktuellen simStep an out ran (3 8chars).*/
  void addSimStep( unsigned char* out, long simStep ) const; 

  /** Extrahiert den simStep aus in (3 8chars). */
  void extractSimStep( const unsigned char* in );

  /** Der Name des eigenen Teams. */
  static Ath98String theTeamname;
  
  /** Pruefsumme des Teamnamens. */
  static char theTeamnameCheckSum[4];

  /** Zeit zu der die Nachricht gesendet wurde. */
  long theTime;

  /** Der Klartext der Nachricht. */
  Ath98String theMessage;

  /** Absender-Info. Das, was sender() zurueckliefert. Der Wert wird nur intern
    von parseMessageString() gesetzt. */
  int theSenderInfo; 
};

#endif
