#include <stdio.h>
#include <string>

#include "Sensors.h"

#include "../basic/Ath98String.h"
#include "../basic/Log.h"
#include "../basic/Timer.h"

#include "../worldmodel/Neck.h"
#include "../worldmodel/PlayerTypes.h"
#include "../worldmodel/WorldModelInterface.h"

Sensors::Sensors( Communicator* aCommunicator )
  : theCommunicator( aCommunicator ) {
    theIdString = "Sensors";
    theSensedFlags = vector<StableObject>(FLAG_COUNT);
    theSensedLines = vector<StableObject>(LINE_COUNT);
    theSensedPlayers = vector<MovingPlayer>(22);
    theSensedBall = MovingObject();
    theSimStep = 0;
    theSimStepOfSoccerServer = 0;
    theHeardMessages = vector<Message>(5);
    theBodyInfos = vector<BodyInfos>(5);
    theTotalCountOfBodyInfos = 0;

    theOpponentGoalieId = UNDEF_PLAYER_ID;
    initSensors();
}

void Sensors::initSensors() {
  theIsNewBodyInfo = NO ;
  theIsNewSeeInfo = NO;
  theIsNewRefereeInfo = NO;
  theIsNewHearInfo = NO;
  theIsNewFullstateInfo=NO;
  theHeardMessages.clear();
  theBodyInfos.clear();
}

void Sensors::sense() {
  Ath98String message;
  static int timeOfLastMessage = globalTimer->getAgentTime();

  try {
    while(1) {
      message = theCommunicator->getNextMessage(0);
      timeOfLastMessage = globalTimer->getAgentTime();
      parseSensoryData(message);
    }
  }
  catch( Communicator::NothingReceived ) {
    int timeSinceLastMessage = globalTimer->getAgentTime() - timeOfLastMessage;
    if ( timeSinceLastMessage > globalConf->communicationTimeout * 1000 ) {
      throwException( SensorsException( "Server has probably terminated." ) );
    }
  }
}

void Sensors::parseSensoryData(const Ath98String& message) {
try {
  if( message.beginsWith( "(see" ) ) {
    theSensedFlags.clear();
    theSensedLines.clear();
    theSensedPlayers.clear();
    theSensedBall.setId(UNDEF_ID);
    char seeMessage[MAX_MESSAGE];

    sscanf( message.cString(), "(see %ld%[^\n]", &theSimStep, seeMessage);
    parseSights(seeMessage);
    theSimStepOfSoccerServer = theSimStep;
    theIsNewSeeInfo = YES;
    globalAnalyzer->haveToAnalyze = YES;
  }
  else if( message.beginsWith( "(hear" ) ) {
    Message heardMessage = Message( message );
/*
cout << heardMessage.time() << "   SENSORS::HEAR" << message << endl << endl;
cout << heardMessage.time() << "   SENSORS::HEAR" << heardMessage << endl << endl;
*/    
    theHeardMessages.push_back( heardMessage );
    theSimStepOfSoccerServer = heardMessage.time();
    if( heardMessage.senderInfo() == REFEREE_ID ) {
      theIsNewRefereeInfo = YES;
    }
    else if( heardMessage.senderInfo() != SELF_ID ){
      theIsNewHearInfo = YES;
    }
  }
  else if( message.beginsWith( "(sense_body" ) ) {
    ++theTotalCountOfBodyInfos;
    theLastBodyInfos = BodyInfos( message );
    theLastBodyInfos.setCount( theTotalCountOfBodyInfos );
    theBodyInfos.push_back( theLastBodyInfos );
    theSimStepOfSoccerServer = theLastBodyInfos.time();
    theIsNewBodyInfo = YES;
  }
  else if( message.beginsWith( "(server_param" ) ) {
    // fngt die neuen vom server 7.0 gesendeten ServerParameter ab
  }
  else if( message.beginsWith( "(player_param" ) ) {
    // fngt die neuen vom server 7.0 gesendeten PlayerParameter ab
  }
  else if( message.beginsWith( "(player_type" ) ) {
    // fngt den neuen vom server 7.0 gesendeten PlayerTyp ab
    globalPlayerTypes->parsePlayerTypeMessage(message);
  }
  else if( message.beginsWith( "(change_player_type" ) ) {
    // fngt den neuen vom server 7.0 gesendeten PlayerTyp ab
    globalPlayerTypes->parseSubstituteMessage(message);
  }
  else if( message.beginsWith( "(fullstate" ) ) {
    char fullstateMessage[MAX_MESSAGE];
    long fullstateSimStep = 0;
    sscanf( message.cString(), "(fullstate %ld%[^\n]", &fullstateSimStep, fullstateMessage );
    theCorrectSituation = Situation( fullstateSimStep );
    parseFullstate( fullstateMessage );
    theIsNewFullstateInfo = YES;
  } 
  else if( message.beginsWith( "(error" ) ) {
    WARNING( Log::SENSORS, "Error message received." );
  }
  else {
    cerr << endl << message << endl;
    cerr << "Sensors::parseSensoryData  >> Unknown answer from the SoccerServer\n";
//    throwException( SensorsException("Unknown answer from the SoccerServer") );
  }
}
catch(...) {
cerr << "Sensors::parseSensoryData  >> schwerer AusnahmeFehler !\n";
}
}


void Sensors::parseSights(char* seeMessage) {
try { 
  double dist, dir;
  double distChange, dirChange;
  double bodyDir;
  double faceDir;
  char objectArray[256];
  char* object;
  char information[30];

  INFO( Log::SENSORS  ,  6  ,  idString() , globalLog->infoStream <<
  " parseSights: " <<    seeMessage  <<  " \n"  ) ;

	agentTimeOfLastSeeInfo = globalTimer->getAgentTime();
//cout << globalWorldModel->simStep() << "  neue Sichtdaten\n";
  while( seeMessage[0] != ')' ) {
    object = objectArray;
    sscanf( seeMessage, " ((%[^)]) %[^)]) %[^\n]", object,
            information, seeMessage);

    INFO( Log::SENSORS  ,  13  ,  idString() , globalLog->infoStream <<
  " parseSights: object: " <<    object  <<
          " <<  \tinformation: " <<    information  <<  " \n"  ) ;


    strcat( object, ")" );
    strcat( information, ")" );

    sscanf( information, "%lf %[^\n]", &dist, information );
    if( information[0] == ')' ) {
      // nur Richtung wurde erhalten, ignorieren
      continue;
    }
    sscanf( information, "%lf %[^\n]", &dir, information );

    /* alle Winkel werden sofort nach dem Parsen in Rad umgewandelt
       und in unser Winkelsystem transformiert
       ( 0 - 2Pi, mathematisch positiver Sinn) */
    dir = normalizeAngle( radFromDegree( -1.0 * dir ) );
    

    distChange = UNDEF_DOUBLE;
    dirChange = UNDEF_DOUBLE;
    bodyDir = UNDEF_DOUBLE;
    faceDir = UNDEF_DOUBLE;

    if( information[0] != ')' ) {
      sscanf( information, "%lf %[^\n]", &distChange, information );
      if( information[0] != ')' ) {
        sscanf( information, "%lf %[^\n]", &dirChange, information );

        dirChange = ( radFromDegree( dirChange ) );
        if( information[0] != ')' ) {
          sscanf( information, "%lf %[^\n]", &bodyDir, information );
          bodyDir = normalizeAngle ( radFromDegree( -1.0 * bodyDir ) );
          if( information[0] != ')' ) {
            sscanf( information, "%lf %[^\n]", &faceDir, information );
            faceDir = normalizeAngle ( radFromDegree( -1.0 * faceDir ) );
          }
        }
      }
    }

    if( strncasecmp (object, "b", 1) == 0 ) { // ball oder Ball
      theSensedBall = MovingObject( BALL_ID, Vector(dist, dir, POLAR, RADIAN),
                      Vector((double)distChange, (double)dirChange, CARTESIAN) );

      INFO( Log::SENSORS  ,  6  ,  idString() , globalLog->infoStream <<
	    " theSensedBall at : "
	    << theSimStepOfSoccerServer
	    << " " << theSensedBall.pos().x() << " " << theSensedBall.pos().y()
	    << " " << theSensedBall.speed().x() << " " << theSensedBall.speed().y()
	    << "\n"; );
//cout << theSimStepOfSoccerServer << "Ball gefunden\n";	    
    }
    else if( strncasecmp(object, "p", 1) == 0 ) { // player oder Player
      char team[256];
      int playerNumber;
      int playerId = UNDEF_PLAYER_ID;
      bool isGoalie = NO;

      if( strncmp(object, "P", 1) == 0 ) { // Player
        sscanf( object, "P%[^\n]", object);
      }
      else { // player
        sscanf( object, "p%[^\n]", object);
      }


      if( object[0] != ')' ) {
        sscanf( object, " \"%[^)\" \n\t\r\f]\"%[^\n]", team, object);
        if( globalWorldModel->teamName() == Ath98String(team) ) {
          playerId = MIN_TEAMMATE_ID;
        }
        else {
          playerId = MIN_OPPONENT_ID;
        }
        if( object[0] != ')' ) {
	        sscanf( object, " %i%[^\n]", &playerNumber, object);
          playerId += playerNumber - 1;
        }
        else {
          if( playerId == MIN_TEAMMATE_ID ) {
            playerId = UNDEF_TEAMMATE_ID;
          }
          if( playerId == MIN_OPPONENT_ID ) {
            playerId = UNDEF_OPPONENT_ID;
          }
        }
        if( object[0] != ')' ) {
	        if( object[1] == 'g' ) { //goalie gefunden
            isGoalie = YES;
            setOpponentGoalieId( playerId );
          }
	      }
      }

      // Point ist relativ zum Spieler, noch nicht absolut
      theSensedPlayers.push_back( MovingPlayer( playerId, isGoalie,
	       Vector(dist, dir, POLAR, RADIAN),
	       Vector(distChange, dirChange, CARTESIAN),
		     faceDir, bodyDir ) );
    }
    else if( strncasecmp(object, "g", 1) == 0 ) { // goal oder Goal
      if( strncmp(object, "g r", 3) == 0) {
        theSensedFlags.push_back(
            StableObject( RIGHT_GOAL_CENTER_FLAG_ID,
                          Vector(dist, dir, POLAR, RADIAN) ) );
      }
      else if( strncmp(object, "g l", 3) == 0) {
        theSensedFlags.push_back(
            StableObject( LEFT_GOAL_CENTER_FLAG_ID,
                          Vector(dist, dir, POLAR, RADIAN) ) );
      }
      else {
        // anonymes Tor, ignorieren.
        continue;
      }
    }
    else if( strncasecmp(object, "f", 1) == 0 ) { // flag oder Flag
      if( strcmp(object, "F)") == 0 ) {
        // anonyme Flagge, ignorieren.
        continue;
      }

      int flagId = UNDEF_FLAG_ID;
      sscanf( object, "f %[^\n]", object);

      switch( object[0] ) {
        case 'l':
          object++;  object++;
          switch( object[0] ) {
            case '0':
              flagId = OUTSIDE_LEFT_CENTER_FLAG_ID;
              break;
            case 't':
              object++;  object++;
              switch( object[0] ) {
                case '\0':
                  flagId = UPPER_LEFT_FLAG_ID;
                  break;
                case '1':
                  flagId = OUTSIDE_LEFT_TOP_10_FLAG_ID;
                  break;
                case '2':
                  flagId = OUTSIDE_LEFT_TOP_20_FLAG_ID;
                  break;
                case '3':
                  flagId = OUTSIDE_LEFT_TOP_30_FLAG_ID;  
                  break;  
              } 
              break;  
            case 'b': 
              object++;  object++;  
              switch( object[0] ) { 
                case '\0': 
                  flagId = LOWER_LEFT_FLAG_ID;  
                  break;  
                case '1': 
                  flagId = OUTSIDE_LEFT_BOTTOM_10_FLAG_ID;  
                  break;  
                case '2': 
                  flagId = OUTSIDE_LEFT_BOTTOM_20_FLAG_ID;  
                  break;  
                case '3': 
                  flagId = OUTSIDE_LEFT_BOTTOM_30_FLAG_ID;  
                  break;  
              } 
              break;  
          } 
          break;  
        case 'r': 
          object++;  object++;  
          switch( object[0] ) { 
            case '0':
              flagId = OUTSIDE_RIGHT_CENTER_FLAG_ID;  
              break;  
            case 't': 
              object++;  object++;  
              switch( object[0] ) { 
                case '\0': 
                  flagId = UPPER_RIGHT_FLAG_ID;  
                  break;  
                case '1': 
                  flagId = OUTSIDE_RIGHT_TOP_10_FLAG_ID;  
                  break;  
                case '2': 
                  flagId = OUTSIDE_RIGHT_TOP_20_FLAG_ID;  
                  break;
                case '3': 
                  flagId = OUTSIDE_RIGHT_TOP_30_FLAG_ID;  
                  break;  
              } 
              break;  
            case 'b': 
              object++;  object++;  
              switch( object[0] ) { 
                case '\0': 
                  flagId = LOWER_RIGHT_FLAG_ID;  
                  break;  
                case '1': 
                  flagId = OUTSIDE_RIGHT_BOTTOM_10_FLAG_ID;  
                  break;  
                case '2': 
                  flagId = OUTSIDE_RIGHT_BOTTOM_20_FLAG_ID;  
                  break;  
                case '3': 
                  flagId = OUTSIDE_RIGHT_BOTTOM_30_FLAG_ID;  
                  break;  
              } 
              break;  
          } 
          break;  
        case 'c': 
          object++;  object++;  
          switch( object[0] ) {
            case '\0': 
              flagId = CENTER_FLAG_ID;  
              break;  
            case 't': 
              flagId = UPPER_MIDDLE_FLAG_ID;  
              break;  
            case 'b': 
              flagId = LOWER_MIDDLE_FLAG_ID;  
              break;  
          } 
          break;  
        case 'p': 
          object++;  object++;  
          switch( object[0] ) {
            case 'l': 
              object++;  object++;  
              switch( object[0] ) { 
                case 't': 
                  flagId = UPPER_LEFT_PENALTY_FLAG_ID;  
                  break;  
                case 'c': 
                  flagId = MIDDLE_LEFT_PENALTY_FLAG_ID;  
                  break;  
                case 'b': 
                  flagId = LOWER_LEFT_PENALTY_FLAG_ID;  
                  break;  
              } 
              break;  
            case 'r': 
              object++;  object++;  
              switch( object[0] ) { 
                case 't': 
                  flagId = UPPER_RIGHT_PENALTY_FLAG_ID;  
                  break;  
                case 'c': 
                  flagId = MIDDLE_RIGHT_PENALTY_FLAG_ID;  
                  break;  
                case 'b': 
                  flagId = LOWER_RIGHT_PENALTY_FLAG_ID;  
                  break;  
              }
              break;  
          } 
          break;  
        case 'g': 
          object++;  object++;  
          switch( object[0] ) { 
            case 'l': 
              object++;  object++;  
              switch( object[0] ) { 
                case 't': 
                  flagId = LEFT_GOAL_TOP_FLAG_ID;  
                  break;  
                case 'b': 
                  flagId = LEFT_GOAL_BOTTOM_FLAG_ID;
                  break;  
              } 
              break;  
            case 'r': 
              object++;  object++;  
              switch( object[0] ) {         
                case 't': 
                  flagId = RIGHT_GOAL_TOP_FLAG_ID;  
                  break;  
                case 'b': 
                  flagId = RIGHT_GOAL_BOTTOM_FLAG_ID;  
                  break;  
              } 
              break;  
          } 
          break;  
        case 't': 
          object++;  object++;  
          switch( object[0] ) { 
            case '0': 
              flagId = OUTSIDE_TOP_CENTER_FLAG_ID;  
              break;  
            case 'l': 
              object++;  object++;  
              switch( object[0] ) { 
                case '1': 
                  flagId = OUTSIDE_TOP_LEFT_10_FLAG_ID;
                  break;  
                case '2': 
                  flagId = OUTSIDE_TOP_LEFT_20_FLAG_ID;  
                  break;  
                case '3': 
                  flagId = OUTSIDE_TOP_LEFT_30_FLAG_ID;  
                  break;  
                case '4': 
                  flagId = OUTSIDE_TOP_LEFT_40_FLAG_ID;  
                  break;  
                case '5': 
                  flagId = OUTSIDE_TOP_LEFT_50_FLAG_ID;  
                  break;  
              }
              break;  
            case 'r' : 
              object++;  object++;  
              switch( object[0] ) { 
                case '1': 
                  flagId = OUTSIDE_TOP_RIGHT_10_FLAG_ID;    
                  break;  
                case '2': 
                  flagId = OUTSIDE_TOP_RIGHT_20_FLAG_ID;  
                  break;  
                case '3': 
                  flagId = OUTSIDE_TOP_RIGHT_30_FLAG_ID;  
                  break;  
                case '4': 
                  flagId = OUTSIDE_TOP_RIGHT_40_FLAG_ID;  
                  break;  
                case '5': 
                  flagId = OUTSIDE_TOP_RIGHT_50_FLAG_ID;  
                  break;              
              } 
              break;  
          } 
          break;  
        case 'b': 
          object++;  object++;  
          switch( object[0] ) { 
            case '0':
              flagId = OUTSIDE_BOTTOM_CENTER_FLAG_ID;  
              break;  
            case 'l': 
              object++;  object++;  
              switch( object[0] ) { 
                case '1': 
                  flagId = OUTSIDE_BOTTOM_LEFT_10_FLAG_ID;  
                  break;  
                case '2': 
                  flagId = OUTSIDE_BOTTOM_LEFT_20_FLAG_ID;  
                  break;  
                case '3': 
                  flagId = OUTSIDE_BOTTOM_LEFT_30_FLAG_ID;  
                  break;
                case '4': 
                  flagId = OUTSIDE_BOTTOM_LEFT_40_FLAG_ID;  
                  break;  
                case '5': 
                  flagId = OUTSIDE_BOTTOM_LEFT_50_FLAG_ID;  
                  break;  
              } 
              break;  
            case 'r' : 
              object++;  object++;  
              switch( object[0] ) { 
                case '1': 
                  flagId = OUTSIDE_BOTTOM_RIGHT_10_FLAG_ID;  
                  break;  
                case '2': 
                  flagId = OUTSIDE_BOTTOM_RIGHT_20_FLAG_ID;  
                  break;  
                case '3': 
                  flagId = OUTSIDE_BOTTOM_RIGHT_30_FLAG_ID;  
                  break;  
                case '4': 
                  flagId = OUTSIDE_BOTTOM_RIGHT_40_FLAG_ID;  
                  break;  
                case '5': 
                  flagId = OUTSIDE_BOTTOM_RIGHT_50_FLAG_ID;  
                  break;              
              }
              break;
          }
          break;
      }

      if(flagId == UNDEF_FLAG_ID) {
        WARNING( Log::SENSORS, "Unknown flag received." );
      }
      else {
        theSensedFlags.push_back(
            StableObject( flagId, Vector(dist, dir, POLAR, RADIAN) ) );
      }
    }
    else if( strncasecmp (object, "l", 1) == 0 ) {
      int lineId = UNDEF_LINE_ID;  
      sscanf( object, "l %[^\n]", object);  
 
      switch( object[0] ) { 
      case 'l': 
        lineId = LEFT_LINE_ID;  
        break;  
      case 'r': 
        lineId = RIGHT_LINE_ID;  
        break;  
      case 't': 
        lineId = UPPER_LINE_ID;  
        break;  
      case 'b': 
        lineId = LOWER_LINE_ID;  
        break;  
      } 
 
      if(lineId == UNDEF_LINE_ID) { 
        WARNING( Log::SENSORS, "Unknown line received." );  
      } 
      else { 
        theSensedLines.push_back(
            StableObject( lineId, Vector(dist, dir, POLAR, RADIAN) ) );  
      } 
    } 
    else {
      Ath98String mesg( "Unknown type of object." );
      mesg = mesg + Ath98String( object );
cerr << "Sensrs::parseSights()  >> Unknown type of object !!!\n";
//      throwException( SensorsException( mesg.cString() ) );  
    } 
  }
}
catch(...) {
cerr << "Sensrs::parseSights()  >> schwerer AusnahmeFehler!\n";
} 
}

void Sensors::parseFullstate(char* fullstateMessage) {
  double x, y, xChange, yChange, bodyDir, faceDir, stamina, effort, recovery;
  char objectArray[256];
  char* object;
  char information[256];

  INFO( Log::SENSORS  ,  6  ,  idString() , globalLog->infoStream <<
  " parseFullstate: " <<    fullstateMessage  <<  " \n"  ) ;

  while( fullstateMessage[0] != ')' ) {
    object = objectArray;
    sscanf( fullstateMessage, " (%[^ ]%[^)]) %[^\n]", object,
            information, fullstateMessage);

    strcat( object, ")" );
    strcat( information, ")" );
    
    if( strncasecmp (object, "p", 1) == 0 ) { // pmode - Playmodeinfo
    
    }
    else if( strncasecmp (object, "v", 1) == 0 ) { // vmode - Viewmodeinfo
    
    }
    else if( strncasecmp (object, "s", 1) == 0 ) { // score - Scoreinfo
    
    }
    else { // Objekt auf dem Spielfeld - Ball oder Spieler

      INFO( Log::SENSORS, 10, idString(), globalLog->infoStream <<
            "Analyzing correct information: " << information << "\n" );
      globalLog->flushInfoStream();

      // parse information string    
      x = UNDEF_DOUBLE;
      y = UNDEF_DOUBLE;
      xChange = UNDEF_DOUBLE;
      yChange = UNDEF_DOUBLE;
      bodyDir = UNDEF_DOUBLE;
      faceDir = UNDEF_DOUBLE;
      stamina = UNDEF_DOUBLE;
      effort = UNDEF_DOUBLE;
      recovery = UNDEF_DOUBLE;
      
      sscanf( information, "%lf %lf %[^\n]", &x, &y, information );
      // Umrechnung Noda-System (rechtsdrehend) auf Karthesisch (linksdrehend)
      y=0-y;
       
      /* Der Spieler erhaelt seine lokalen Informationen stets so, als wuerde er
         von links nach rechts spielen. Daher ist hier fuer die absoluten
         Koordinaten eine Umrechnung notwendig */   
      if( globalWorldModel->side() == RIGHT_SIDE_ID ) {
        x = 0-x; y = 0-y;
      }
      if( information[0] != ')' ) {
        sscanf( information, "%lf %lf %[^\n]", &xChange, &yChange, information );
        yChange = 0-yChange;
        if( globalWorldModel->side() == RIGHT_SIDE_ID ) {
          xChange = 0-xChange; yChange = 0-yChange;
        }
        if( information[0] != ')' ) {
          sscanf( information, "%lf %lf %[^\n]", &bodyDir, &faceDir, 
                  information );
          bodyDir = radFromDegree( 0-bodyDir );
          faceDir = radFromDegree( 0-faceDir );
          if( globalWorldModel->side() == RIGHT_SIDE_ID ) {
            bodyDir =  bodyDir - Ath98_PI ;
            faceDir = faceDir - Ath98_PI ;
          }
          bodyDir = normalizeAngle( bodyDir );
          faceDir = bodyDir + faceDir;
          faceDir = normalizeAngle( faceDir );        
        }
        if( information[0] != ')' ) {
          sscanf( information, "%lf %lf %lf", &stamina, &effort, 
                  &recovery );
        }
      }

      INFO( Log::SENSORS, 10, idString(), globalLog->infoStream <<
            "Analyzing correct object: " << object << "\n" );
      globalLog->flushInfoStream();

      // parse object string    
      if( strncasecmp (object, "b", 1) == 0 ) { // ball oder Ball
        theCorrectSituation.setBall( ModelledBall( BALL_ID, Vector( x, y ),
                       Vector( xChange, yChange ),
                       UNDEF_TEAM_ID, -1,
                       Confidence(theCorrectSituation.simStep(), 0.0, 1.0),
                       Confidence(theCorrectSituation.simStep(), 0.0, 1.0) ) );
      }
      else if( strncasecmp(object, "l", 1) == 0 ||
               strncasecmp(object, "r", 1) == 0 ) { // player oder Player
        char side;
        int sideID;
        int playerNumber;
        int playerId = UNDEF_PLAYER_ID;
        bool isGoalie = NO;

        sscanf( object, " %c_%i", &side, &playerNumber);
        sideID=((side == 'l') ? LEFT_SIDE_ID : RIGHT_SIDE_ID);
        if( globalWorldModel->side() == sideID ) {
            playerId = MIN_TEAMMATE_ID + playerNumber - 1;
        }
        else {
          playerId = MIN_OPPONENT_ID + playerNumber - 1;
        }
          
        theCorrectSituation.addPlayer( ModelledPlayer( playerId, Vector( x, y ),
            Vector( xChange, yChange ), faceDir, bodyDir, VIEW_WIDTH_WIDE, 
            VIEW_QUALITY_HIGH, stamina, effort, recovery,
            isGoalie, 
            Confidence(theCorrectSituation.simStep(), 0.0, 1.0),
            Confidence(theCorrectSituation.simStep(), 0.0, 1.0),
            Confidence(theCorrectSituation.simStep(), 0.0, 1.0),
            Confidence(theCorrectSituation.simStep(), 0.0, 1.0),
            Confidence(theCorrectSituation.simStep(), 0.0, 1.0),
            Confidence(theCorrectSituation.simStep(), 0.0, 1.0),
            Confidence(theCorrectSituation.simStep(), 0.0, 1.0) ) );
      }
      
    }
    
    
  }

}
