#ifndef _SENSORS_H_
#define _SENSORS_H_

#include <string>
#include "BodyInfos.h"
#include "Communicator.h"
#include "Message.h"
#include "../basic/Exception.h"
#include "../worldmodel/Situation.h"
#include "../worldmodel/StableObject.h"
#include "../worldmodel/MovingObject.h"
#include "../worldmodel/MovingPlayer.h"

/** Das Sensorik-Objekt. 
  Die Sensorik wartet auf Sensorinformationen vom Server
  und erzeugt eine Objekt-Repraesentation von ihnen. Diese Sensorinformationen
  werden dann vom Weltmodell ausgelesen.

  Bereits bei der Erzeugung der Objektdarstellung der Wahrnehmung wird vom
  Server- in das mathematische Koordinatensystem gewandelt.

  Von der Klasse wird nur ein Objekt erzeugt. Die Klasse ist nicht zur Vererbung
  geeignet.

  {\small {\bf Quelltext}

  [\URL[Header]{src/Sensors.h} | [\URL[Implementierung]{src/Sensors.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: Sensors.h,v $
  *   $Revision: 1.1.1.1 $
  *   $Date: 2001/07/31 22:23:09 $
  *   $Author: rami $ 
  *
  **************************************************
  \end{verbatim} }*/
class Sensors
{
public:

  /** Interne Fehlerklasse. */
  class SensorsException: public Exception {
  public:
    SensorsException(const Ath98String& aDescription )
      : Exception( aDescription ) {}
  };
  
  /** Erzeugt das Sensorobjekt.
    @param aCommunicator Das Communicator-Objekt, von dem die 
    Sensor-Informationen kommen.  */
  Sensors( Communicator* aCommunicator );
		
  /** Initialisieren der Sensordaten. Die Arrays der K"orper- und 
    H"orinformationen werden
    gel"oscht, die booleschen Variablen, die aussagen ob neue Sicht-
    bzw. Schiedsrichterinformationen vorliegen, werdem auf NO zur"uckgesetzt.
    Somit steht das Sensorobjekt f"ur den Empfang neuer
    Daten bereit. Diese Methode muss nach dem kompletten Lesen der
    aufgelaufenen
    Daten aufgerufen werden um damit kenntlich zu machen, dass man nur noch
    an neueren Daten interessiert ist. */
  void initSensors();

  /** Empfaengt Sensorstrings und verarbeitet sie zu Objekten.
    Versucht vom eigenen Communicator Ath98Strings zu lesen. Diese werden zu
    Objekten verarbeitet und spaeter vom Weltmodell ausgelesen.

    Nach einem Aufruf von Sense liegen keine Ath98Strings mehr im Communicator
    bereit. 
    @exception SensorsException wenn der Server l\"anger als
    \Ref{Configuration::communicationTimeout}() Sekunden keine
    Sichtinformationen geschickt hat. Man kann dann davon ausgehen, da\3 der
    Server sich terminiert hat. */
  void sense();

  /** Ein Ath98String vom SoccerServer wird geparst. Die zu diesem Ath98String
    gehoerenden Objekte werden aufgebaut. */
  void parseSensoryData(const Ath98String& message);

  /** Anfrage ob eine neue Bodyinformation vorliegt. Falls dies der Fall
    ist wird YES zurueckgegeben, andernfalls NO. Diese Methode dient dem
    Timer und dem Weltmodell um zu erkennen, ob sich die Welt geaendert hat.
    Neue Daten liegen vor, wenn nach dem letzten \Ref{initSensors} Sensordaten 
    ausgewertet wurden. */
  bool isNewBodyInfo();

 /** Anfrage ob eine neue Sichtinformation vorliegt. Falls dies der Fall
    ist wird YES zurueckgegeben, andernfalls NO. Diese Methode dient dem
    Timer und dem Weltmodell um zu erkennen, ob sich die Welt geaendert hat.
    Neue Daten liegen vor, wenn nach dem letzten \Ref{initSensors} Sensordaten
    ausgewertet wurden. */
  bool isNewSeeInfo();

  /** Anfrage ob eine neue Schiedsrichterentscheidungen vorliegen. Falls
      dies der Fall ist wird YES zurueckgegeben, andernfalls NO. Diese
      Methode dient dem Weltmodell um zu erkennen, ob sich die Welt
      geaendert hat. Neue Daten liegen vor, wenn nach dem letzten
      \Ref{initSensors} Schiedsrichterdaten ausgewertet wurden. */
  bool isNewRefereeInfo();

  /** Anfrage ob eine neue H\"orinformation vorliegt, die keine
      Schiedsrichterentscheidung ist. Falls
      dies der Fall ist wird YES zurueckgegeben, andernfalls NO. Diese
      Methode dient dem Weltmodell um zu erkennen, ob sich die Welt
      geaendert hat. Neue Daten liegen vor, wenn nach dem letzten
      \Ref{initSensors} SchiedsrichterdatenH\"ordaten ausgewertet wurden. */ 
  bool isNewHearInfo();

  /** Anfrage ob eine neue Fullstateinformation vorliegt. Falls
      dies der Fall ist wird YES zurueckgegeben, andernfalls NO. Diese
      Methode dient dem Timer um zu erkennen, ob neue korrekte Informationen
      in das Weltmodell uebernommen werden muessen. Neue Fullstateinformationen
      werden in \Ref{isNewInfo} nicht beruecksichtigt, da die normale
      Abarbeitung im Agenten durch die zusaetzlichen Informationen, die zudem
      nur waehrend der Testphase vorliegen, nicht beeinflut werden soll.
      Neue Daten liegen vor, wenn nach dem letzten
      \Ref{initSensors} Fullstatedaten ausgewertet wurden. */ 
  bool isNewFullstateInfo();

  /** Anfrage ob eine neue Information vorliegt (Body, See, Refree oder Hear).
      Falls dies der Fall ist wird YES zurueckgegeben, andernfalls NO. Diese
      Methode dient dem Timer und dem Weltmodell um zu erkennen, ob sich die
      Welt geaendert hat. */
  bool isNewInfo();


  /** Liefert ein Array mit den wahrgenommenen Flaggen. Zu diesen gehoeren 
    auch die Tore. Die Informationen entstammen der see-Nachricht des
    SoccerServers. */
  const vector<StableObject>& sensedFlags() const;
  
  /** Liefert ein Array mit den wahrgenommenen Linien. Die Position der Linien
    ist besonders zu interpretieren: Die Laenge ist die Entfernung zwischen der
    eigenen Position und dem Schnittpunkt der eigenen Blickgeraden und der
    Linie. Der Winkel ist der Winkel der Linie relativ zur eigenen
    Blickrichtung. Die Informationen zu den Linien entstammen der
    see-Nchricht des SoccerServers. */
  const vector<StableObject>& sensedLines() const;
  
  /** Die gesehenen Spieler. Diese entstammen der see-Nachricht des
    SoccerServers. */
  const vector<MovingPlayer>& sensedPlayers() const;
  
  /** Der gesehene Ball. Falls keiner gesehen wurde, wird eine 
    NothingSeen Exception ausgeloest. Der Ball wird der see-Nachricht
    des SoccerServers entnommen. */
  const MovingObject& sensedBall() const;
  
  /** Der Simulationsschritt zu der die Sichtinformationen gehoeren. 
    Diese Angabe gilt fuer sensedFlags, sensedLines, sensedPlayers und 
    sensedBall. */
  long simStep();

  /** Der Simulationsschritt in der der sich der SoccerServer befindet.
    Dieser Simulationsschritt basiert auf der letzten mitbekommenden
    Zeit (see, bodyInfos oder Referee ). */
  long simStepOfSoccerServer();

  /** Die gehoerten Nachrichten. Diese entstammen der hear-Nachricht des
    SoccerServers. */
  const vector<Message>& heardMessages() const;

  /** Die Informationen ueber den eigenen Koerper. Diese entstammen der 
    sense_body-Nachricht des SoccerServers. */
  const vector<BodyInfos>& bodyInfos() const;
  
  /** Die letzten Informationen \"uber den eigenen K\"orper. */
  const BodyInfos& lastBodyInfos() const;

  /** Rueckgabe der durch fullstate gelieferten korrekten Situation. */
  const Situation& correctSituation() const;

  /** Liefert die Id des gegnerischen Goalies. */
  const int opponentGoalieId() const;

  /** Setzt die Id des gegnerischen Goalies. */
  void setOpponentGoalieId( int anId );

  /** Die Id des gegnerischen Goalies - Hack */
  int theOpponentGoalieId;

  /** Liefert einen String mit Klassenbezeichnung zurueck. */
  string idString() const;

  int agentTimeOfLastSeeInfo;

protected:

  /** String der die Kurzkennung der Klasse enthlt */
  string theIdString;

private:
  /** Interne Referenz auf Communicator Objekt */
  Communicator *theCommunicator;

  /** Die gesehenen Flaggen inclusive Toren. */
  vector<StableObject> theSensedFlags;

  /** Die gesehenen Linien. */
  vector<StableObject> theSensedLines;

  /** Die gesehenen Spieler. */
  vector<MovingPlayer> theSensedPlayers;

  /** Der gesehene Ball. */
  MovingObject theSensedBall;

  /** Der Simulationsschritt, zu dem die Sichtinformationen gehoeren. */
  long theSimStep;

  /** Der Simulationsschritt des SoccerServers. */
  long theSimStepOfSoccerServer;

  /** Die gehoerten und aufbereiteten Nachrichten. */
  vector<Message> theHeardMessages;

  /** Die Informationen ueber den eigenen Koerper. */
  vector<BodyInfos> theBodyInfos;

  /** Die letzten Informationen \"uber den eigenen K\"orper. */
  BodyInfos theLastBodyInfos;

  /** Die korrekte Sichtinformation als Situation (nur fuer fullstate). */
  Situation theCorrectSituation;

  /** Anzahl der seit dem Start des Spieles empfangenen BodyInfos. */
  long int theTotalCountOfBodyInfos;

 /** Liegt neue Sichtinformation vor? */
  bool theIsNewBodyInfo;

  /** Liegt neue Sichtinformation vor? */
  bool theIsNewSeeInfo;

  /** Liegen neue Schiederichterentscheidungen vor? */
  bool theIsNewRefereeInfo;

  /** Liegt neue H\"orinformation vor. */
  bool theIsNewHearInfo;
  
  /** Liegen neue Fullstateinformation vor. */
  bool theIsNewFullstateInfo;

  /** Parsen einer Sichtinformation. */
  void parseSights(char* seeMessage);

  /** Parsen einer Sichtinformation. */
  void parseFullstate(char* seeMessage);
};

inline const vector<StableObject>& Sensors::sensedFlags() const {
  return theSensedFlags;
}

inline const vector<StableObject>& Sensors::sensedLines() const {
  return theSensedLines;
}

inline const vector<MovingPlayer>& Sensors::sensedPlayers() const {
  return theSensedPlayers;
}

inline const MovingObject& Sensors::sensedBall() const {
  return theSensedBall;
}

inline long Sensors::simStep() {
  return theSimStep;
}

inline long Sensors::simStepOfSoccerServer() {
  return theSimStepOfSoccerServer;
}

inline const vector<Message>& Sensors::heardMessages() const {
  return theHeardMessages;
}

inline const vector<BodyInfos>& Sensors::bodyInfos() const {
  return theBodyInfos;
}

inline const BodyInfos& Sensors::lastBodyInfos() const {
  return theLastBodyInfos;
}

inline bool Sensors::isNewBodyInfo() {
  return theIsNewBodyInfo;
}

inline const Situation& Sensors::correctSituation() const {
  return theCorrectSituation;
}

inline const int Sensors::opponentGoalieId() const{
  return theOpponentGoalieId;
}

inline void Sensors::setOpponentGoalieId( int anId ){
  if( (anId >= MIN_OPPONENT_ID) && (anId <= MAX_EXISTING_OPPONENT_ID) ){
    if( (anId != theOpponentGoalieId) && (theOpponentGoalieId != UNDEF_PLAYER_ID) ){
      WARNING( Log::SENSORS, "Received GoalieId is different from previous one." );
    }
    theOpponentGoalieId = anId;
  }
}

inline bool Sensors::isNewSeeInfo() {
  return theIsNewSeeInfo;
}

inline bool Sensors::isNewRefereeInfo() {
  return theIsNewRefereeInfo;
}

inline bool Sensors::isNewHearInfo(){
  return theIsNewHearInfo;
}

inline bool Sensors::isNewFullstateInfo(){
  return theIsNewFullstateInfo;
}

inline bool Sensors::isNewInfo(){
  return (theIsNewBodyInfo || theIsNewSeeInfo ||
	  theIsNewRefereeInfo || theIsNewHearInfo);
}


inline string Sensors::idString() const
{
  return theIdString;
}
#endif

