#ifndef _CONFIGUREDBALL_H_
#define _CONFIGUREDBALL_H_

#include <stream.h>
#include "ModelledObject.h"
#include "ModelledPlayer.h"
#include "MovingObject.h"

#include "../basic/Common.h"
#include "../basic/Configuration.h"
#include "../basic/Vector.h"

 
/** 
 * Aktueller Stand 
 * 
 *   $RCSfile: ConfiguredBall.h,v $ 
 *   $Revision: 1.1.1.1 $ 
 *   $Date: 2001/07/31 22:23:09 $ 
 *   $Author: rami $  
 * 
 */ 

class ConfiguredBall : public ModelledObject
{
public:
  /** @name Interne Fehlerklassen. */
  //@{
    /** Allgemeine Fehlerklasse. */
    class ConfiguredBallException: public ModelledObjectException {
    public:
      ConfiguredBallException(const Ath98String& aDescription )
        : ModelledObjectException( aDescription ) {}
    }; 
    
    /** Fehlerklasse f\"ur illegalen Kickversuch.
      Sie wird ausgel\"ost, wenn der Ball au\3er Reichweite des Spielers 
      ist. */
    class KickImpossibleException: public ConfiguredBallException {
    public:
      KickImpossibleException(const Ath98String& aDescription )
        : ConfiguredBallException( aDescription ) {}
    };

    //@}


    ConfiguredBall(  int id  = UNDEF_ID,
		     const Vector& aPos = Vector(), 
		     const Vector& aSpeed = Vector(),
		     const Confidence& aPosConfidence = Confidence(),
	  	     const Confidence& aSpeedConfidence = Confidence() );

  /** Ausgabe auf einen String-Stream */
    friend ostream& operator<< (ostream&, const ConfiguredBall& );

  /** kompakte Ausgabe auf einen Stream fr constant Log */
    void toStampStream( ostream& os) const;


    /** Es wird die tatsaechlich verfuegbare Kraft
      (effectiveKickPower) ausgerechnet.
      @param power Die Kraft, mit der man den Ball kicken will.
      @param kickingPlayer Der Spieler der die Beschleunigung ausfuehrt.
      @return effectiveKickPower Die Kraft, mit der der Ball tatsaechlich
      beschleunigt wird.
      @exception KickImpossibleException falls Ball nicht im Kick-Bereich (zu 
      weit weg) */
    double effectiveKickPower( double power,
			   const ModelledPlayer& kickingPlayer ) const;

    /** Es wird die Kraft bestimmt, die dem SoccerServer geschickt werden
      muss, damit effektiv der Ball mit der Kraft power beschleunigt wird.
      Diese Funktion ist also genau die Umkehrfunktion von effectiveKickPower.
      Es gilt also: kickPowerForSend(effectiveKickPower( power )) == power,
      wenn power normalisiert ist (im definierten Bereich liegt). Liegt das
      zu sendene Kommando nicht im definierten Bereich, wird normalisiert
      und eine Warnung ins Log geschrieben.
      @param power Die Kraft mit der Ball beschleunigt werden soll.
      @param kickingPlayer Der Spieler der die Beschleunigung ausfuehrt.
      @return powerForSend Die Kraft mit dem man den Ball kicken muss.
      @exception KickImpossibleException falls  Ball nicht im Kick-Bereich (zu 
      weit weg) */
    double kickPowerForSend( double power,
			     const ModelledPlayer& kickingPlayer ) const;

    /** Liefert den maximalen sinnvollen effektiven Kick durch den der 
      angegebene Spieler den Ball zur derzeitigen Situation beschleunigen kann. */
    double maxEffectiveKickPower(const ModelledPlayer& kickingPlayer) const;

    /** Normalisiert die Kraft fuer einen kick.
      @exception ConfiguredBallException falls der uebergebene Wert UNDEF_DOUBLE 
      ist */
    double normalizeKickPower(double power) const;

    /** Normalisiert den Winkel fuer einen kick. */
    double normalizeKickAngle(double angle) const;
  //@}

    
  /** @name Konfigurierbare Werte erfragen */
  //@{
    /** Klassenvariablen BallSize abfragen */
    double size() const;

    /** Klassenvariablen BallDecay abfragen */
    double decay() const;

    /** Klassenvariablen BallMaxSpeed abfragen */
    double maxSpeed() const;

    /** Klassenvariablen KickPowerRate abfragen */
    double kickPowerRate() const;

    /** Klassenvariablen MaxPower abfragen */
    double maxPower() const;

    /** Klassenvariablen MinPower abfragen */
    double minPower() const;

    /** Klassenvariablen MaxMoment abfragen */
    double maxMoment() const;

    /** Klassenvariablen MinMoment abfragen */
    double minMoment() const;

    /** Klassenvariablen KickMargin abfragen */
    double kickMargin() const;

    /** Klassenvariablen BallSize setzen */
    void setSize(double aBallSize);

    /** Klassenvariablen BallDecay setzen */
    void setDecay(double aBallDecay);

    /** Klassenvariablen BallMaxSpeed setzen */
    void setMaxSpeed(double aBallMaxSpeed);

    /** Klassenvariablen KickPowerRate setzen */
    void setKickPowerRate(double aKickPowerRate);

    /** Klassenvariablen MaxPower setzen */
    void setMaxPower(double aMaxPower);

    /** Klassenvariablen MinPower setzen */
    void setMinPower(double aMinPower);

    /** Klassenvariablen MaxMoment setzen */
    void setMaxMoment(double aMaxMoment);

    /** Klassenvariablen MinMoment setzen */
    void setMinMoment(double aMinMoment);

    /** Klassenvariablen KickMargin setzen */
    void setKickMargin(double aKickMargin);
  //@}

  
  /** @name Konfigurierbare Werte setzen */
  //@{ 
    /** liest die Werte aller Klassenvariablen aus dem Konfigurationsfile 
    ein. */
    void setClassVars();
  //@}

private:
  ///
  static double theBallSize;
  ///
  static double theBallDecay;
  ///
  static double theBallMaxSpeed;
  ///
  static double theKickPowerRate;
  ///
  static double theMaxPower;
  ///
  static double theMinPower;
  ///
  static double theMaxMoment;
  ///
  static double theMinMoment;
  ///
  static double theKickMargin;

protected:
  ///
  static short  isSetOnce;

};

inline double ConfiguredBall::normalizeKickPower( double power ) const {
  if( power == UNDEF_DOUBLE ) {
    throwException( ConfiguredBallException( "undef power value" ) );
  }
  else if( power < minPower() ) {
    return minPower();
  }
  else if( power > maxPower() ){
    return maxPower();
  }
  else {
    return power;
  }
}


inline double ConfiguredBall::normalizeKickAngle( double angle ) const {
  angle = normalizeAngle(angle);
  
  if( (angle <= Ath98_PI) && (angle > normalizeAngle(-1.0*minMoment()) ) ) {
    return normalizeAngle( -1.0*minMoment() );
  }
  else if( (angle > Ath98_PI) && (angle < normalizeAngle(-1.0*maxMoment()) ) ) {
    return normalizeAngle( -1.0*maxMoment() );
  }
  else {
    return angle;
  }
}


inline double ConfiguredBall::size() const {
  return ConfiguredBall::theBallSize;
}

inline double ConfiguredBall::decay() const {
  return ConfiguredBall::theBallDecay;
}

inline double ConfiguredBall::maxSpeed() const {
  return ConfiguredBall::theBallMaxSpeed;
}

inline double ConfiguredBall::kickPowerRate() const {
  return ConfiguredBall::theKickPowerRate;
}

inline double ConfiguredBall::maxPower() const {
  return ConfiguredBall::theMaxPower;
}

inline double ConfiguredBall::minPower() const {
  return ConfiguredBall::theMinPower;
}

inline double ConfiguredBall::maxMoment() const {
  return ConfiguredBall::theMaxMoment;
}

inline double ConfiguredBall::minMoment() const {
  return ConfiguredBall::theMinMoment;
}

inline double ConfiguredBall::kickMargin() const {
  return ConfiguredBall::theKickMargin;
}

inline void ConfiguredBall::setSize(double aBallSize) {
  ConfiguredBall::theBallSize = aBallSize;
}

inline void ConfiguredBall::setDecay(double aBallDecay) {
  ConfiguredBall::theBallDecay=aBallDecay;
}

inline void ConfiguredBall::setMaxSpeed(double aBallMaxSpeed) {
  ConfiguredBall::theBallMaxSpeed=aBallMaxSpeed;
}

inline void ConfiguredBall::setKickPowerRate(double aKickPowerRate) {
  ConfiguredBall::theKickPowerRate=aKickPowerRate;
}

inline void ConfiguredBall::setMaxPower(double aMaxPower) {
  ConfiguredBall::theMaxPower=aMaxPower;
}

inline void ConfiguredBall::setMinPower(double aMinPower) {
  ConfiguredBall::theMinPower=aMinPower;
}

inline void ConfiguredBall::setMaxMoment(double aMaxMoment) {
  ConfiguredBall::theMaxMoment=aMaxMoment;
}

inline void ConfiguredBall::setMinMoment(double aMinMoment) {
  ConfiguredBall::theMinMoment=aMinMoment;
}

inline void ConfiguredBall::setKickMargin(double aKickMargin) {
  ConfiguredBall::theKickMargin = aKickMargin;
}

inline ostream& operator<< (ostream& os, const ConfiguredBall& ball )
{
  os << 
"( ID: " << ball.id()  <<
", Pos: " << ball.pos() << 
", Speed: " << ball.speed() << 
" )" ;
  return os;
}

inline void ConfiguredBall::toStampStream( ostream& os) const {
  os << 
    " " << pos().x() << 
    " " << pos().y() <<
    " " << speed().x() <<
    " " << speed().y();
}
#endif
