#ifndef _FIELD_H_
#define _FIELD_H_

#include <string>

#include "StableObject.h"

#include "../basic/Common.h"
#include "../basic/Exception.h"
#include "../basic/Vector.h"


/**
  Kapselung des Wissens ueber das Layout des Spielfeldes. Hier koennen absolute
  Positionen von Flaggen, Toren und Linien erfragt werden. Ausserdem findet
  sich hier die Berechnung eines Blickpunktes aus einer
  Reihe von relativen Positionsangaben. 

  Sinn der Klasse ist es, die Daten eines Fussballfeldes wegzukapseln.

  {\small {\bf Quelltext}

  [\URL[Header]{src/Field.h} | [\URL[Implementierung]{src/Field.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: Field.h,v $
  *   $Revision: 1.1.1.1 $
  *   $Date: 2001/07/31 22:23:09 $
  *   $Author: rami $
  *
  **************************************************
  \end{verbatim} }*/
class Field
{
public:

  /** Erzeugt ein neues Feld. Bei der Erzeugung wird die aktuelle Breite
    des Tores aus dem globalen Configuration-Objekt gelesen und die
    Positionen der entsprechenden Flaggen neu gesetzt. */
  Field();

  /** Blickrichtung kann nicht berechnet werden. */
  class NoFaceDirException: public Exception {
  public:
    NoFaceDirException()
      : Exception() {}
  };

  /** Position kann nicht berechnet werden. */
  class NoPosException: public Exception {
  public:
    NoPosException()
      : Exception() {}
  };

  /** Berechnet die Position des Agenten auf dem Spielfeld.
      Das verwendete Verfahren ist Gradient Descent. */
  Vector calculatePosition( const vector<StableObject>& seenFlags ) const;

  /** Liefert YES, wenn sich der Punkt innerhalb des Spielfeldes befindet. */
  static bool isInsideField(const Vector& aPoint);

  /** Gibt die Position eines Punktes zurueck. Ein Punkt kann entweder eine
    Flagge oder ein Tor sein. Weil die Positionen einzelner Torflaggen vor
    dem Spiel noch nicht feststehen (Torbreite wird aus server.conf ausgelesen)
    ist diese Methode keine Klassenmethode.
    @param anId Die Id eines Punktes (wie in Gemeinsame Deklarationen)
    @return Vektor, der die absolute Position des Punktes auf dem Spielfeld
    beschreibt.
    @exception InvalidIdException wenn die Id keine
    existierende Flagge/Tor beschreibt. */
  const Vector& posOfPoint(int anId) const;

  /** Gibt die Laenge des Feldes zurueck.  */
  static double fieldLength();

  /** Gibt die Breite des Feldes zurueck.  */
  static double fieldWidth();

  /** Gibt die Breite des Tores zurueck, diese wird im Konstrukter aus
    dem Configuration-Objekt ausgelessen und damit aktualisiert. */
  double goalWidth() const;

  /** Gibt die Laenge des Strafraums zurueck.  */
  static double penaltyAreaLength();

  /** Gibt die Breite des Strafraums zurueck.  */
  static double penaltyAreaWidth();

  /** Gibt den Radius des Anstosskreis zurueck. */
  static double kickoffCircleRadius();

  /** Liefert die Position des Mittelpunktes des gegenerischen Tores. */
  const Vector& enemyGoal() const;

  /** Liefert die Position des oberen Pfosten des gegnerischen Tores. */
  const Vector& enemyGoalTop() const;

  /** Liefert die Position des unteren Pfosten des gegnerischen Tores. */
  const Vector& enemyGoalBottom() const;

  /** Liefert die Position des Mittelpunktes des eignen Tores. */
  const Vector& ourGoal() const;

  /** Liefert die Position des oberen Pfosten des eignen Tores. */
  const Vector& ourGoalTop() const;

  /** Liefert die Position des unteren Pfosten des eignen Tores. */
  const Vector& ourGoalBottom() const;

  /** Ist der Ball im Strafraum ? */
  bool isPositionInPenaltyArea(const Vector& aVector) const;

  /** Berechnet die absolute Blickrichtung. Es wird eine der uebergebenen
    Linien ausgewaehlt, um daraus den Blickwinkel zu bestimmen.
    @param seenLines Array mit den gesehenen Linien.
    @exception NoFaceDirException Wird erzeugt, wenn keine Linie
    gesehen wurde, oder wenn sich aus der ausgewaehlten Linie der Blickwinkel
    nicht eindeutig ergibt. (Das ist in einem selten auftretenden Spezialfall
    fuer die rechte Linie der Fall.) */
  static double computeFaceDir( const vector<StableObject>& seenLines );

  /** Liefert einen String mit Klassenbezeichnung zurueck. */
  string idString() const;
  
private:

  /** Berechnet aus der relativen Position einer mit ihrer Id gegebenen Flage
    und dem absoluten Blickwinkel, die eigene absolute Position. */
  Vector computePosFromFlag( const StableObject& flag, double faceDir ) const;


  // jetzt die internen Klassenvariablen;

  /** C-Array der Positionen aller Flaggen. Als Index kann FlagId - MINFLAG
    verwendet werden. Das Array hat FLAGCOUNT Eintraege. */
  static Vector theFlagPositions[];

  /** C-Array der Positionen der Tore. Als Index kann GoalId - MINGOAL
    verwendet werden. Das Array hat GOALCOUNT Eintraege. */
  static const Vector theGoalPositions[];

  /** Die Feldlaenge */
  static const double theFieldLength;

  /** Die Feldbreite */
  static const double theFieldWidth;

  /** Die Torbreite, diese wird aus den Configuration-Objekt aktualisiert */
  static double theGoalWidth;

  /** Die Strafraumlaenge */
  static const double thePenaltyAreaLength;

  /** Die Strafraumbreite */
  static const double thePenaltyAreaWidth;

  /** Der Radius des Anstosskreises */
  static const double theKickoffCircleRadius;

  /** String der die Kurzkennung der Klasse enthlt */
  string theIdString;
};

// Die Implementierung der Inline-Funktionen

inline double Field::fieldLength()
{
  return theFieldLength;
}

inline double Field::fieldWidth()
{
  return theFieldWidth;
}

inline double Field::goalWidth() const
{
  return theGoalWidth;
}

inline double Field::penaltyAreaLength()
{
  return thePenaltyAreaLength;
}

inline double Field::penaltyAreaWidth()
{
  return thePenaltyAreaWidth;
}

inline double Field::kickoffCircleRadius()
{
  return theKickoffCircleRadius;
}

inline string Field::idString() const
{
  return theIdString;
}

inline const Vector& Field::enemyGoal() const
{
  return posOfPoint( RIGHT_GOAL_CENTER_FLAG_ID );
}

inline const Vector& Field::enemyGoalTop() const
{
  return posOfPoint( RIGHT_GOAL_TOP_FLAG_ID );
}

inline const Vector& Field::enemyGoalBottom() const
{
  return posOfPoint( RIGHT_GOAL_BOTTOM_FLAG_ID );
}

inline const Vector& Field::ourGoal() const
{
  return posOfPoint( LEFT_GOAL_CENTER_FLAG_ID );
}

inline const Vector& Field::ourGoalTop() const
{
  return posOfPoint( LEFT_GOAL_TOP_FLAG_ID );
}

inline const Vector& Field::ourGoalBottom() const
{
  return posOfPoint( LEFT_GOAL_BOTTOM_FLAG_ID );
}

inline bool Field::isPositionInPenaltyArea(const Vector& aVector) const
{
  if( aVector.x() <= posOfPoint(UPPER_LEFT_PENALTY_FLAG_ID).x() ) {
    if( ( aVector.y() <= posOfPoint(UPPER_LEFT_PENALTY_FLAG_ID).y() ) &&
	( aVector.y() >= posOfPoint(LOWER_LEFT_PENALTY_FLAG_ID).y() ) ) {
      return YES;
    }
  }
  return NO;
}

#endif



