#ifndef _MODELLEDOBJECT_H_
#define _MODELLEDOBJECT_H_

#include "MovingObject.h"
#include "Confidence.h"
#include "../basic/Common.h"
#include "../basic/Exception.h"
#include <math.h>

/**
  Eigene Annahmen ueber sich bewegendes Objekt in der Aussenwelt.
  ModelledObject ist eine abstrakte Superklasse. Moegliche Objekttypen sind
  Spieler und Ball.

  Ein ModelledObject ist die interne Repraesentation eines Objektes der
  Aussenwelt im Weltmodell. Diese Repraesentation ist immer in ABSOLUTEN
  Koordinaten, d.h. pos() gibt die Position des Objektes relativ zur
  Spielfeldmitte an. Die positive x-Achse zeigt in Richtung gegnerisches Tor,
  die y-Achse ist um 90 Grad (im mathematischen Sinn, also gegen den
  Uhrzeigersinn) gedreht. Ebenso ist speed() absolut zu interpretieren. 
  Der Wert
  gibt an, um wieviel sich die Position innerhalb eines Taktes ver\"andert.
  
  Zusaetzlich zu den Eigenschaften eines MovingObject
  hat ein ModelledObject Simulationsmoeglichkeiten, um seine eigene Position
  und
  Geschwindigkeit in der Zukunft zu bestimmen. Dazu besitzt es einen
  Verlangsamungsfaktor und Verlaesslichkeitsinformationen.

  {\small {\bf Quelltext}

  [\URL[Header]{src/ModelledObject.h} | [\URL[Implementierung]{src/ModelledObject.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  *************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: ModelledObject.h,v $
  *   $Revision: 1.1.1.1 $
  *   $Date: 2001/07/31 22:23:09 $
  *   $Author: rami $ 
  *
  **************************************************
  \end{verbatim} }*/
class ModelledObject : public MovingObject
{
public:

  /** @name Interne Fehlerklassen. */
  //@{
    /** Allgemeine Fehlerklasse. Wird von [FEHLT] ausgel\"ost. */
    class ModelledObjectException: public Exception {
    public:
      ModelledObjectException(const Ath98String& aDescription )
        : Exception( aDescription ) {}
    };
  //@}

  /** @name Werte abfragen */
  //@{
    /** Liefert Verlaesslichkeit der Position. */
    const Confidence& posConfidence() const;

    /** Liefert Verlaesslichkeit der Geschwindigkeit. */
    const Confidence& speedConfidence() const;
 
  //@}
    
    
  /** @name Werte setzen */
  //@{
    /** Setzt Verlaesslichkeit der Position. */
    void setPosConfidence( const Confidence& newPosConfidence );

    /** Setzt Verlaesslichkeit der Geschwindigkeit. */
    void setSpeedConfidence( const Confidence& newSpeedConfidence );

  //@}
    
protected:
  /** Protected Default-Konstruktor.
    Die Klasse ist abstrakt, aber der Konstruktor kann von den Nachkommen 
    genutzt werden. */
  ModelledObject();

  /** Neues ModelledObject mit Id, Position, Geschwindigkeit,
  Positions- und Geschwindikgeitsverl\"asslichkeit.
    @param id Die Objekt-Id. M\"ogliche Werte sind in den \Ref{Objekt-Ids}
      aufgelistet.
    @param newPosition Die Position.
    @param newSpeed Die Geschwindigkeit. 
    @param aPosConfidence Die Verlaesslichkeit der Position. Als Default wird 
    RANDOM angenommen.
    @param aPosConfidence Die Verlaesslichkeit der Geschwindigkeit. Als Default
    wird RANDOM angenommen. */
  ModelledObject( int id = UNDEF_ID,
		  const Vector& aPos = Vector(),
		  const Vector& aSpeed = Vector(),
		  const Confidence& aPosConfidence = Confidence(),
		  const Confidence& aSpeedConfidence = Confidence() );

  /** Die Verlaesslichkeit der Position. Sie haengt von der anfaenglichen
    Verlaesslichkeit und von der Zahl der durchgefuehrten Simulationsschritte
    ab. */
  Confidence thePosConfidence;
  
  /** Die Verlaesslichkeit der Geschwindigkeit. Sie haengt von der
    anfaenglichen Verlaesslichkeit und von der Zahl der durchgefuehrten
    Simulationsschritte ab. Dies ist auch gleichzeitig
    die Verl\"asslichkeit f\"ur die Blickrichtung. */
  Confidence theSpeedConfidence;

};


inline const Confidence& ModelledObject::posConfidence() const 
{
  return thePosConfidence;
}

inline const Confidence& ModelledObject::speedConfidence() const 
{
  return theSpeedConfidence;
}

inline void ModelledObject::setPosConfidence(
    const Confidence& newPosConfidence) 
{
  thePosConfidence=newPosConfidence;
}

inline void ModelledObject::setSpeedConfidence(
    const Confidence& newSpeedConfidence) 
{
  theSpeedConfidence=newSpeedConfidence;
}
 
#endif






