#ifndef _MODELLEDPLAYER_H_
#define _MODELLEDPLAYER_H_

#include <stream.h>
#include "ConfiguredPlayer.h"
#include "Confidence.h"
#include "../basic/Configuration.h"
#include "../basic/Common.h"
#include "../basic/Exception.h"
#include "../basic/Vector.h"



/**
  Die interne Repraesentation eines Spielers im Weltmodell.
  Ein ModelledPlayer hat zusaetzlich zu den Eigenschaften eines ModelledObject
  noch eine Blickrichtung und eine Kraftreserve. Die Kraftreserve setzt sich
  aus
  den Werten stamina, effort und recovery zusammen. Dazu kommen entsprechende
  Verl\"asslichkeitsinformationen.

  {\small {\bf Quelltext}

  [\URL[Header]{src/ModelledPlayer.h} | [\URL[Implementierung]{src/ModelledPlayer.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: ModelledPlayer.h,v $
  *   $Revision: 1.1.1.1 $
  *   $Date: 2001/07/31 22:23:09 $
  *   $Author: rami $ 
  *
  **************************************************
  \end{verbatim} }*/
class ModelledPlayer : public ConfiguredPlayer
{

public:

  /** @name Interne Fehlerklassen. */
  //@{
    /** Allgemeine Fehlerklasse. */
    class ModelledPlayerException: public ModelledObjectException {
    public:
      ModelledPlayerException(const Ath98String& aDescription )
        : ModelledObjectException( aDescription ) {}
    };
  //@}
 
       
       
  /** @name Erzeugen und Zerst\"oren */
  //@{
    /** Neuer Spieler mit Daten und Default-Konstruktor.
      @param isGoalie Handelt es sich um den Goalie ?
      @param aPos Die Position. Als Default wird unbekannt angenommen.
      @param aSpeed Die Geschwindigkeit. Als Default wird unbekannt angenommen.
      @param Confidence
      @param aPosConfidence Die Verlaesslichkeit der Position. Als Default wird 
      RANDOM angenommen.
      @param aSpeedConfidence Die Verlaesslichkeit der Geschwindikeit. Als
      Default wird RANDOM angenommen.
      @param aFaceDir Die Blickrichtung. Als Default wird unbekannt angenommen.
      @param aBodyDir Die Laufrichtung. Als Default wird unbekannt angenommen.
      @param aFaceDirConfidence Die Verl\"a\3lichkeit der Blickrichtung. Als
      Default wird RANDOM angenommen.
      @param aBodyDirConfidence Die Verl\"a\3lichkeit der Laufrichtung. Als
      Default wird RANDOM angenommen.
      @param aViewWidth Sichtbreite darf nur die Werte VIEW_WIDTH_NARROW,
      VIEW_WIDTH_NORMAL oder VIEW_WIDTH_WIDE annehmen. Als Default wird 
      VIEW_WIDTH_WIDE angenommen.
      @param stamina die Kraftreserve des Spielers. Gueltige Werte: 0 bis
      theStaminaMax. Kleinere Werte werden auf 0 gesetzt, groessere Werte
      (UNDEF_DOUBLE) auf theStaminaMax.
      @param effort der Leistungsfaktor des Spielers. Gueltige Werte:
      theEffortMin bis 1.0. Kleinere Werte werden auf theEffortMin gesetzt,
      groessere Werte (UNDEF_DOUBLE) auf 1.0.
      @param recovery der Erholungsfactor des Spielers. Gueltige Werte
      theRecoveryMin bis 1.0. Kleinere Werte werden auf theRecoveryMin gesetzt,
      groessere Werte (UNDEF_DOUBLE) auf 1.0.
      @param aStaminaConfidence Die Verl\"asslichkeit der Stamina-Werte. Als
      Default wird RANDOM angenommen.
      @param id Die Id des Spielers
      Default wird NO angenommen */
    
    ModelledPlayer( int aId = UNDEF_ID,
      const Vector& aPos = Vector(), 
      const Vector& aSpeed = Vector(),
      double aFaceDir = 0.0,
      double aBodyDir = 0.0,
      int aViewWidth = VIEW_WIDTH_WIDE,
      int aViewQuality = VIEW_QUALITY_HIGH,
      double stamina = globalConf->playerStaminaMax,
      double effort = 1.0,
      double recovery = 1.0,
      bool isGoalie = NO,
      const Confidence& aPosConfidence = Confidence(),
      const Confidence& aSpeedConfidence = Confidence(),
      const Confidence& aFaceDirConfidence = Confidence(),
      const Confidence& aBodyDirConfidence = Confidence(),
      const Confidence& aStaminaConfidence = Confidence(),
      const Confidence& aTeamIdConfidence = Confidence(),
      const Confidence& aPlayerIdConfidence = Confidence() );
    
    /** Modellierung eines wahrgenommenen Spielers.
      @param aMovingPlayer Der gesehene Spieler.
      @param aPosConfidence Die Verlaesslichkeit der Position. Als Default wird 
      RANDOM angenommen.
      @param aSpeedConfidence Die Verlaesslichkeit der Geschwindikeit. Als
      Default wird RANDOM angenommen.
      @param aFaceDirConfidence Die Verl\"a\3lichkeit der Blickrichtung. Als
      Default wird RANDOM angenommen.
      @param aBodyDirConfidence Die Verl\"a\3lichkeit der Blickrichtung. Als
      Default wird RANDOM angenommen.
      @param aViewWidth Sichtbreite darf nur die Werte VIEW_WIDTH_NARROW,
      VIEW_WIDTH_NORMAL oder VIEW_WIDTH_WIDE annehmen. Als Default wird 
      VIEW_WIDTH_WIDE angenommen.
      @param stamina die Kraftreserve des Spielers. Gueltige Werte: 0 bis
      theStaminaMax. Kleinere Werte werden auf 0 gesetzt, groessere Werte
      (UNDEF_DOUBLE) auf theStaminaMax.
      @param effort der Leistungsfaktor des Spielers. Gueltige Werte:
      theEffortMin bis 1.0. Kleinere Werte werden auf theStaminaMin gesetzt,
      groessere Werte (UNDEF_DOUBLE) auf 1.0.
      @param recovery der Erholungsfactor des Spielers. Gueltige Werte
      theRecoveryMin bis 1.0. Kleinere Werte werden auf theRecoveryMin gesetzt,
      groessere Werte (UNDEF_DOUBLE) auf 1.0.
      @param aStaminaConfidence Die Verl\"asslichkeit der Stamina-Werte. Als
      Default wird RANDOM angenommen.*/
    ModelledPlayer( const MovingPlayer& aMovingPlayer,
      int aViewWidth = VIEW_WIDTH_WIDE,
      int aViewQuality = VIEW_QUALITY_HIGH,
      double stamina = globalConf->playerStaminaMax,
      double effort = 1.0,
      double recovery = 1.0,
      const Confidence& aPosConfidence = Confidence(),
      const Confidence& aSpeedConfidence = Confidence(),
      const Confidence& aFaceDirConfidence = Confidence(),
      const Confidence& aBodyDirConfidence = Confidence(),
      const Confidence& aStaminaConfidence = Confidence(),
      const Confidence& aTeamIdConfidence = Confidence(),
      const Confidence& aPlayerIdConfidence = Confidence() );

  /** Legt eine Kopie des Objekts auf dem Heap an und liefert einen Zeiger
    darauf zurueck. */
  StableObject* copy() const;

  /** @name Werte abfragen */
  //@{
    /** Liefert Verl\"a\3lichkeit der Stamina-Werte. */
    const Confidence& staminaConfidence() const;

    /** Liefert Torwartrolle  */
    bool ModelledPlayer::isGoalie() const;

    /** Liefert Blickrichtung */
    double faceDir() const;

    /** Liefert Krperrichtung */
    double bodyDir() const;

    /** Der Winkel zwischen BodyDir und FaceDir im Bogenmass.
        gemessen von BodyDir nach FaceDir -> 
         Kopf nach links =>  0 < relativeFaceDir < 90
         Kopf nach rechts => 270 < relativeFaceDir < 360 */
    double relativeFaceDir() const;

    /** Liefert die Verl\"a\3lichkeit der Blickrichtung. */
    const Confidence& faceDirConfidence() const;

    /** Liefert die Verl\"a\3lichkeit der Blickrichtung. */
    const Confidence& bodyDirConfidence() const;

    /** Liefert die Verl\"a\3lichkeit der TeamId. */
    const Confidence& teamIdConfidence() const;

    /** Liefert die Verl\"a\3lichkeit der PlayerId. */
    const Confidence& playerIdConfidence() const;

    /** Liefert die Groesse des Gesichtsfeldes.
	@return VIEW_WIDTH_NARROW, VIEW_WIDTH_NORMAL oder VIEW_WIDTH_WIDE */
    int viewWidth() const;

    /** Liefert die Qualitt der Sichtinfos.*/
    int viewQuality() const;


    /** Liefert den Wert des Winkels des sichtbaren Bereiches. */
    double visibleAngle() const;

    /** Liefert die TeamId des Teams, zu dem der Spieler geh\"ort.
      @return OUR_TEAM_ID, THEIR_TEAM_ID oder UNDEF_TEAM_ID
    */
    int team() const;



    /** Liefert YES wenn die Nummer des Spielers bekannt ist, sonst NO. */
    bool isPlayerNumberKnown() const;

    /** Beschreibung als Ath98String fuer die Ausgabe auf einem Terminal.
      Es werden ID, Position, Geschwindigkeit, Zuverlaessigkeitswerte, FaceDir,
      Sichtrichtung und -winkel (in Grad), sowie die Staminawerte
      ausgegeben. */
    Ath98String toString() const;

  /** Ausgabe auf einen String-Stream */
    friend ostream& operator<< (ostream&, const ModelledPlayer& );

  /** kompakte Ausgabe auf einen Stream fr constant Log */
    void toStampStream( ostream& os) const;
  //@}


  /** @name Werte setzen */
  //@{
    /** Setzt die Goalie-Eigenschaft */
    void setIsGoalie( bool isGoalie );

    /** Setzt die Verl\"a\3lichkeit der Stamina-Werte. */
    void setStaminaConfidence( const Confidence& newStaminaConfidence );


    /** Setzt Blickrichtung neu. */
    void setFaceDir( double newFaceDir );

    /** Setzt Verl\"a\3lichkeit der Blickrichtung. */
    void setFaceDirConfidence( const Confidence& newFaceDirConfidence );

    /** Setzt Laufrichtung neu. */
    void setBodyDir( double newFaceDir );

    /** Setzt Verl\"a\3lichkeit der Laufrichtung. */
    void setBodyDirConfidence( const Confidence& newFaceDirConfidence );

    /** Setzt Verl\"a\3lichkeit der TeamId. */
    void setTeamIdConfidence( const Confidence& newTeamIdConfidence );

    /** Setzt Verl\"a\3lichkeit der PlayerId. */
    void setPlayerIdConfidence( const Confidence& newPlayerIdConfidence );

   /** Setzt die Groesse des Gesichtsfeldes. 
      Es wird nicht auf Gueltigkeit getestet.
      @param newViewWidth Neue Gesichtsfeldgroesse. Moegliche Werte sind 
      VIEW_WIDTH_NARROW, VIEW_WIDTH_NORMAL, VIEW_WIDTH_WIDE. */
    void setViewWidth( int newViewWidth );

   /** Setzt die Qualitaet der Sichtinfos. 
      Es wird nicht auf Gueltigkeit getestet.
      @param newViewQuality  Neue Sichtqualitaet. Moegliche Werte sind 
      VIEW_QUALITY_LOW, VIEW_QUALITY_HIGH. */
    void setViewQuality( int newViewQuality );

    /** Setzt den Sichtwinkel. Dieser ist abhaengig vom Paramter viewWidth und
      von den normalen Sichtwinkel der aus der Datei server.conf entnommen wird.
      @param viewWidth Sichtbreite darf nur die Werte VIEW_WIDTH_NARROW,
      VIEW_WIDTH_NORMAL oder VIEW_WIDTH_WIDE annehmen.
      @exception ModelledPlayerException bei undefinierter viewWidth */
    void setVisibleAngle( int viewWidth );  
  //@}

    /** Liefert den maximalen sinnvollen effektiven Dash mit dem der Spieler
	zur derzeitigen Situation beschleunigen kann.
	@exception ModelledPlayerException bei undefinierter Geschwindigkeit */
    double maxEffectiveDashPower() const;
     
  /** @name Simulation */

  //@{
    /** Simuliert das Verstreichen von Simulationszeit. Die Routine geht davon
      aus, dass der Spieler waehrend dieser Zeit keine Aktionen unternimmt. Es
      wird simuliert: Die Geschwindigkeitsaenderung, die Positionsaenderung und
      die Staminaaenderung. Die Verlaesslichkeit wird herabgesetzt.
      @param stepCount Zahl der Simulationsschritte. */
    void simulateSteps( unsigned int simSteps = 1 );


    /** Beschleunigt. Wie stark sich die Kraft auf die Beschleunigung auswirkt,
      haengt vom Kraftfaktor (powerRate) und einem Objektbeschleunigungsfaktor
      (maxPowerRate) ab. Dabei wird auch die Kraftreserve herabgesetzt.
      @param aPower Die Kraft, mit der beschleunigt wird. 
      @exception ModelledPlayerException bei undefinierter Geschwindigkeit */
    void accelerate( double aPower );

    /** Dreht die Blick- und Laufrichtung. Die naechste Beschleunigung geht 
      standardmaessig in diese Richtung.
      @param angle Ein Winkel relativ zur bisherigen Blickrichtung.  */
    void turn( double angle );
    
    /** Dreht nur den Nacken des Spielers.
      @param angle Ein Winkel im 2001-Winkelsystem relativ zur bisherigen Blickrichtung.  */
    void turnNeck( double angle );
  //@}
    
    
  /** @name Winkelbezogene Hilfsfunktionen */
  //@{
    /** Ermittelt den Winkel unter dem der Spieler den angegebenen Punkt sieht.
      @exception ModelledPlayerException bei undefinierter Blickrichtung */
    double angleOfFaceTo( const Vector& point ) const;

   /** Ermittelt den Winkel unter dem der Spieler den angegebenen Punkt sehen 
       wuerde, wenn er geradeaus schaut.
      @exception ModelledPlayerException bei undefinierter Blickrichtung */
    double angleOfBodyTo( const Vector& point ) const;

    /** Ermittelt ob sich das angegeben Objekt im Sichtbereich des
      Spielers befindet. Wenn sich das Objekt im Sichtbereich befindet
      wird YES zurueckgegeben, andernfalls NO. */
    bool isInFieldOfView( const StableObject& pos ) const;  

    /** Ermittelt ob sich das angegeben Objekt in unmittelbarer N\"ahe des
      Spielers befindet. Wenn sich das Objekt in unmittelbarer N\"ahe des
      Spielers befindet (der Spieler das Objekt f\"uhlen kann)
      wird YES zurueckgegeben, andernfalls NO. */
    bool isInFieldOfFeel( const StableObject& pos ) const;  

    /** Ermittelt ob sich das angegeben Objekt im Wahrnehmungsbereich des
      Spielers befindet. Wenn sich das Objekt im Wahrnehmungsbereich befindet
      wird YES zurueckgegeben, andernfalls NO. Ein Objekt befindet sich im
      Wahrnehmungsbereich des Spielers, wenn es entweder in dessen
      Sichtbereich ist, oder in unmittelbarer N\"ahe des Spielers ist
      @see isInFieldOfView
      @see isInFieldOfFeel */
    bool isInFieldOfSense( const StableObject& pos ) const;
    
  //@}  

  /** @name Entfernunsberechnungen */
  //@{
    /** Gibt den Abstand des Spielers von der angegebenen Position zurueck.
      Dieser Abstand wird vom Mittelpunkt des Spielers gemessen.
      @param position Die Position zu der Abstand ermittelt werden soll.
      @return distance Abstand des Spielers zur Position position. */
    double distTo( const Vector& position ) const;

  //@}    


  /***/
  double getMaxDistAfterSteps(int steps, int offset=0) const;

 
  /// Maximale Schrittzahl die ein Spieler braucht um zum Ball zu gelangen.
  static const int MAX_STEPS_TO_BALL;

  
private:
 
  // alle Winkel im 2001-Format ( 0 < angle < 360) !!
  
  /** Die Blickrichtung als absoluter Winkel im Bogenmass. */
  double theFaceDir;

  /** Die Krperrichtung als absoluter Winkel im Bogenmass. */
  double theBodyDir;

  /** Die Verl\"a\3lichkeit der Blickrichtung. Sie haengt von der anfaenglichen
    Verlaesslichkeit und von der Zahl der durchgefuehrten Simulationsschritte
    ab. */
  Confidence theFaceDirConfidence;

  /** Die Verl\"a\3lichkeit der Blickrichtung. Sie haengt von der anfaenglichen
    Verlaesslichkeit und von der Zahl der durchgefuehrten Simulationsschritte
    ab. */
  Confidence theBodyDirConfidence;

  /** Die Verl\"a\3lichkeit der Stamina-Werte. Sie haengt von der anfaenglichen
    Verlaesslichkeit und von der Zahl der durchgefuehrten Simulationsschritte
    ab. */
  Confidence theStaminaConfidence;

  /** Die Verlsslichkeit der Teamzugehrigkeit  */
  Confidence theTeamIdConfidence;
  
  /** Die Verlsslichkeit der PlayerId  */
  Confidence thePlayerIdConfidence;
  
  /** Die der aktuellen Sichtfeldgroesse entsprechende Konstante */
  int theViewWidth;

  /** Die der aktuellen Sichtqualitaet entsprechende Konstante */
  int theViewQuality;

  /** Der sichtbare Bereich als Winkel im Bogenmass. */
  double theVisibleAngle;


  /** Die Goalie-Eigenschaft. */
  bool isTheGoalie;

  

  /** Formel mit der der Drehwinkel modifiziert wird */
  double turnAngleMult() const;

  /** Rechnet die Koordinaten des Spielers relativ zum Ball und dessen Richtung aus. */
  Vector newCoordinates(const Vector &ballPos, const Vector &playerPos, const Vector &ballSpeed) const;

};



inline double ModelledPlayer::distTo( const Vector& position ) const
{
  return ( (position - pos()  ).length() );
}


inline bool ModelledPlayer::isGoalie() const
{
  return ((isTheGoalie) || (id() == 420));
}

inline double ModelledPlayer::faceDir() const 
{
  return theFaceDir;
}

inline double ModelledPlayer::bodyDir() const 
{
  return theBodyDir;
}

inline double ModelledPlayer::relativeFaceDir() const 
{
  // liefert im Normalfall einen Wert zwischen -90 (Kopf nach links)
  //  und +90 (Kopf nach rechts)
  double relFaceDir = bodyDir() - faceDir();

  // relFaceDir > 180 --> quatorsprung (berspringen der 0/360 grad-Grenze in der Berechnung)
  //  und Kopf nach links
  if( relFaceDir > Ath98_PI ){
    relFaceDir -= (2.0 * Ath98_PI);
  }
  else{
    // relFaceDir < -180 --> quatorsprung (berspringen der 0/360 grad-Grenze in der Berechnung)
    //  und Kopf nach rechts
    if( relFaceDir < (-1.0 * Ath98_PI) ){
      relFaceDir += (2.0 * Ath98_PI);
    }
  }
  // Umwandlung in 2001-Winkelsystem
  return (relFaceDir >= 0.0 + DEFAULT_COMPARISON_ACCURACY ) ? (2.0 * Ath98_PI) - relFaceDir : (-1.0 * relFaceDir);
}

inline void ModelledPlayer::setIsGoalie( bool isGoalie )   
{
  isTheGoalie=isGoalie;
}
 
inline void ModelledPlayer::setFaceDir( double newFaceDir ) 
{
  theFaceDir = normalizeAngle( newFaceDir );
}

inline void ModelledPlayer::setBodyDir( double newBodyDir ) 
{
  theBodyDir = normalizeAngle( newBodyDir );
}

inline void ModelledPlayer::turnNeck( double angle )  
{   
  double relAngle = ( angle > Ath98_PI ) ? (2.0 * Ath98_PI) - angle : (-1.0 * angle);
  setFaceDir( normalizeAngle( faceDir() - relAngle ) );
} 
 
inline void ModelledPlayer::setViewWidth( int newViewWidth ) 
{ 
  theViewWidth = newViewWidth;  
  setVisibleAngle( newViewWidth );  
} 

inline void ModelledPlayer::setViewQuality( int newViewQuality ) 
{ 
  theViewQuality = newViewQuality;  
} 


inline int ModelledPlayer::viewWidth() const 
{
  return theViewWidth;
}

inline int ModelledPlayer::viewQuality() const 
{
  return theViewQuality;
}

inline double ModelledPlayer::visibleAngle() const 
{
  return theVisibleAngle;
}


inline const Confidence& ModelledPlayer::staminaConfidence() const
{
  return theStaminaConfidence;
}

inline void ModelledPlayer::setStaminaConfidence(
    const Confidence& newStaminaConfidence) 
                               {
  theStaminaConfidence = newStaminaConfidence;
}

inline const Confidence& ModelledPlayer::faceDirConfidence() const
{
  return theFaceDirConfidence;
}

inline const Confidence& ModelledPlayer::bodyDirConfidence() const 
{
  return theBodyDirConfidence;
}

inline void ModelledPlayer::setFaceDirConfidence(
    const Confidence& newFaceDirConfidence) 
{
  theFaceDirConfidence = newFaceDirConfidence;
}

inline void ModelledPlayer::setBodyDirConfidence(
    const Confidence& newBodyDirConfidence) 
{
  theBodyDirConfidence = newBodyDirConfidence;
}

inline const Confidence& ModelledPlayer::playerIdConfidence() const
{
  return thePlayerIdConfidence;
}

inline void ModelledPlayer::setPlayerIdConfidence(
    const Confidence& newPlayerIdConfidence) {
  thePlayerIdConfidence = newPlayerIdConfidence;
}

inline const Confidence& ModelledPlayer::teamIdConfidence() const
{
  return theTeamIdConfidence;
}

inline void ModelledPlayer::setTeamIdConfidence(
    const Confidence& newTeamIdConfidence) {
  theTeamIdConfidence = newTeamIdConfidence;
}

inline ostream& operator<< (ostream& os, const ModelledPlayer& player )
{
  os << 
    "( ID: "  <<   player.id()  << 
    ", Id" << player.playerIdConfidence() <<
    ", Team" << player.teamIdConfidence() <<
    ", IsGoalie: " << player.isGoalie() <<
    ", Pos:" << player.pos()  <<
    ", Pos"  << player.posConfidence()  <<
    ", Speed:" << player.speed()  <<
    ", Speed"  << player.speedConfidence()  <<
    ", FaceDir: " <<  degFromRadiant( player.faceDir() ) << 
    ", FaceDir"  << player.faceDirConfidence()  <<
    ", BodyDir: " <<  degFromRadiant( player.bodyDir() ) << 
    ", BodyDir"  << player.bodyDirConfidence()  <<
    ", VisibleAngle: " << degFromRadiant( player.visibleAngle() ) <<
    ", Stamina: " <<  player.stamina()  <<
    ", Recovery: " <<  player.recovery()  <<
    ", Effort: " <<  player.effort()  <<
    ", StaminaValues"  << player.staminaConfidence()  <<
    " )" ;
  return os;
}

inline void ModelledPlayer::toStampStream( ostream& os) const {
  os << 
    " " << MODELLED_PLAYER_MARK << 
    " " << id()  << 
    " " << isGoalie() <<
    " " << pos().x() << 
    " " << pos().y() <<
    " " << posConfidence().reliability() <<
    " " << speed().x() <<
    " " << speed().y() <<
    " " << speedConfidence().reliability() <<
    " " << bodyDir();
}


inline double ModelledPlayer::getMaxDistAfterSteps(int steps, int offset) const{
  return globalPlayerTypes->getMaxDistForPlayerAfterStep(thePlayerType, steps, offset);
}

#endif


