#ifndef _NECK_H_
#define _NECK_H_

#include <string>
#include "WorldModelInterface.h"
#include "../basic/Action.h"
#include "../serverInterface/BodyInfos.h"
#include "../serverInterface/Sensors.h"

/**
  Der Nacken des Agenten. 
 
  Die Klasse zur Steuerung des Nackens ist ein kleiner Zustandsautomat, der von
  den anderen Komponenten des Agenten weitgehend abgekoppelt ist. Sie wird von 
  den Sensoren ueber eingehende Sicht- und Koerperinformationen informiert. Die 
  Optionen koennen sich ein bestimmtes Verhalten des Nackens wuenschen. Anhand
  dieser Informationen werden Kommandos zur Drehung des Nackens generiert. Diese
  werden nicht in den Plan eingefuegt, den eine Option generiert, sondern von
  den Effektoren regelmaessig bei der Klasse erfragt und gesendet. Dadurch ist
  es moeglich, die Kommandos immer zwischen zwei Sichtinformationen zu senden. 

  
  {\small {\bf Quelltext}

  [\URL[Header]{src/Neck.h} | [\URL[Implementierung]{src/Neck.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: Neck.h,v $
  *   $Revision: 1.1.1.1 $
  *   $Date: 2001/07/31 22:23:09 $
  *   $Author: rami $ 
  *
  **************************************************
  \end{verbatim} }*/
class Neck
{
 public:

  /** Minimale Drehung, die ausgefuehrt werden soll. */
  // static const double MIN_TURN_ANGLE_IN_DEGREE = 1.0;

  /** Konstruktor */
  Neck();

  /** Setzt das Verhalten des Nackens, die Ausrichtung des Neckens bzgl. dem
      Objekt und die Prioritaet der Aktion. Jede Option kann das gewuenschte
      Verhalten des Nackens festlegen */
  void setNeckBehaviour( const int desiredNeckBehaviour = NECK_IN_FRONT,
			 const int desiredNeckBehaviourAlignment = ALIGN_CENTER, const int
			 desiredNeckBehaviourPriority = VERY_LOW_PRIORITY);

  /** ... es soll ein Spieler angesehen werden (NECK_TO_PLAYER_WITH_ID) */
  void setNeckBehaviour( const int desiredNeckBehaviour, const int
			 desiredNeckBehaviourAlignment, const int desiredNeckBehaviourPriority,
			 const int desiredId);

  /** ... es soll eine Position angesehen werden (NECK_TO_POS) */
  void setNeckBehaviour( const int desiredNeckBehaviour, const int
			 desiredNeckBehaviourAlignment, const int desiredNeckBehaviourPriority,
			 const Vector desiredPos);

  /** Setzt die Qualitaet der Sicht */
  void setViewBehaviour( const int desiredViewWidth, const int
			 desiredViewQuality, const int desiredViewPriority );

  /** Nach dem Absenden eines Kommandos wird der Nacken informiert */
  void registerSendedAction( const Action& anAction );

  /** Berechnet abhaengig vom gewuenschten Verhalten das naechste Kommando */
  void generateNeckCommand();

  /** Berechnet abhaengig von der gewuenschten Sicht das naechste Kommando */
  void generateViewCommand();

  /** Die Effektoren erfragen regelmaessig, ob ein Kommando zur Drehung des
      Nackens gesendet werden muss */
  bool needToSendNeckCommand();

  /** Die Effektoren erfragen regelmaessig, ob ein Kommando zum Aendern der
      Sicht gesendet werden muss */
  bool needToSendViewCommand();
  
  /** Die Effektoren erfragen regelmaessig, ob ein Kommando zur Drehung des
      Halses gesendet werden muss */
  Action getTurnNeckCommand();

  /** Die Effektoren erfragen regelmaessig, ob ein Kommando zum Aendern der
      Sicht gesendet werden muss */
  Action getChangeViewCommand();
  
  /** Ermittelt den zulaessigen Winkel aus dem gewuenschten Winkel
      desiredAngle:  absoluter Winkel zwischen player und desiredObject
      Rckgabewert: normalisierter Winkel fr Drehung mit maximal mglicher Sicht auf das Objekt*/
  double angleAcceptingLimitedNeckPosition( const double desiredAngle );

  /** Liefert einen String mit Klassenbezeichnung zurueck. */
  string idString() const;

  /** Fragt das Verhalten des Nackens ab */
  int getNeckBehaviour();

  /** Fragt die Ausrichtung des Nackens bzgl. dem Objekt ab */
  int getNeckBehaviourAlignment();

  /** Fragt die Prioritaet des Verhaltens des Nackens ab */
  int getNeckBehaviourPriority();

  /** Fragt die Id des Spielers der bei NECK_TO_PLAYER_WITH_ID angesehen werden
      soll. */
  int getTheIdOfThePlayerToBeSeen();

  /** Fragt die Position die bei NECK_TO_POS angesehen werden
      soll. */
  Vector getThePosToBeSeen();

  /** Berechnung des Halsdrehwinkels, um eine Position unter Beruecksichtung
      der gewuenschten Ausrichtung im Blick zu haben. */
  double headAngleToTurnTo( const Vector& pos ) const;

  protected :
    
    /** String der die Kurzkennung der Klasse enthaelt */
    string theIdString;

 private:
  /** Ist der \"ubergebene Winkel gro\3 genug, damit ein turn_neck
      generiert werden soll. Die Toleranz ist abh\"angig vom Sichtwinkel:
      je kleiner der Sichtwinkel und je n\"aher am Rand des Sichtbereichs
      geguckt wird, desto geringer die Toleranz.*/
  bool isAngleGreatEnough( double aAngle );

  /** die relative Richtung zwischen FaceDir und BodyDir
      (zwischen 0 und 90 (links) sowie 270 und 360 (rechts)) */
  double relativeFaceDir();


  /** Wie soll der Nacken sich verhalten */ 
  int theNeckBehaviour; 

  /** Wie soll der Nacken ausgerichtet werden */ 
  int theNeckBehaviourAlignment; 

  /** Mit welcher Prioritaet soll der Nacken gedreht werden */ 
  int theNeckBehaviourPriority; 

  /** Auf welche Breite soll die Sicht geaendert werden */
  int theViewWidth;

  /** Auf welche Qualitaet soll die Sicht geaendert werden */
  int theViewQuality;

  /** Mit welcher Prioritaet soll die Sicht geaendert werden */ 
  int theViewPriority;

  /** Flag, das bestimmt, ob ein Befehl fuer den Nacken generiert werden muss */
  bool theNeedToSendNeckCommand;

  /** Flag, das bestimmt, ob ein Befehl fuer die Sicht generiert werden muss */
  bool theNeedToSendViewCommand;

  /** Die Id des Spielers der bei einer NECK_TO_PLAYER_WITH_ID Aktion angesehen
      werden soll. */
  int theIdOfThePlayerToBeSeen;

  /** Die Position die bei einer NECK_TO_POS Aktion angesehen
      werden soll. */
  Vector thePosToBeSeen;

  /** Liefert die gewuenschte Aktion fuer den Nacken */
  Action theTurnNeckCommand;

  /** Liefert die gewuenschte Aktion fuer die Sicht */
  Action theChangeViewCommand;

  double theMinNeckAngle;
  double theMaxNeckAngle;

  /** der aktuelle Sichtwinkel */
  int currentViewWidth() const;

  /** der aktuelle Sichtwinkel */
  int currentViewQuality() const;

  /** die aktuelle Koerperrichtung */
  double bodyDir() const;

  /** die aktuelle Blickrichtung */
  double faceDir() const;

};



/** Global verfuegbarer Nacken */
extern Neck *globalNeck;


inline string Neck::idString() const
{
  return theIdString;
}

inline void Neck::registerSendedAction( const Action& anAction)
{
  if (anAction.id() != SENSE_BODY_ACTION) {
    INFO( Log::NECK  ,  3  ,  idString() , globalLog->infoStream
      << "NECK::registerAction" << anAction << endl);  
  } 
}


inline bool Neck::needToSendViewCommand()
{ 
  generateViewCommand();
  
  if (theNeedToSendViewCommand==YES) {
    INFO( Log::NECK  ,  3  ,  idString() , globalLog->infoStream << 
	  "NECK::needToSendViewCommand(YES) \n");
  } 
  else {  
    INFO( Log::NECK  ,  3  ,  idString() , globalLog->infoStream << 
	  "NECK::needToSendViewCommand(NO) \n"); 
  }

  return theNeedToSendViewCommand;
}


inline bool Neck::needToSendNeckCommand()
{ 
  generateNeckCommand();
  
  if (theNeedToSendNeckCommand==YES) {
    INFO( Log::NECK  ,  3  ,  idString() , globalLog->infoStream << 
	  "NECK::needToSendNeckCommand(YES) \n");
  } 
  else {  
    INFO( Log::NECK  ,  3  ,  idString() , globalLog->infoStream << 
	  "NECK::needToSendNeckCommand(NO) \n"); 
  }
// HACK !  
//  return theNeedToSendNeckCommand;
  return YES;
}


inline Action Neck::getChangeViewCommand()
{
  INFO( Log::NECK  ,  3  ,  idString() , globalLog->infoStream << 
	"NECK::getChangeViewCommand() \n");  

  theNeedToSendViewCommand = NO;
  theViewPriority = NO_PRIORITY;

  return theChangeViewCommand;
}


inline Action Neck::getTurnNeckCommand()
{
  INFO( Log::NECK  ,  3  ,  idString() , globalLog->infoStream << 
	"NECK::getTurnNeckCommand() \n");  

  theNeedToSendNeckCommand = NO;
  theNeckBehaviourPriority = NO_PRIORITY;

  return theTurnNeckCommand;
}

inline int Neck::getNeckBehaviour()
{
  return theNeckBehaviour;
}


inline int Neck::getNeckBehaviourAlignment()
{
  return theNeckBehaviourAlignment;
}


inline int Neck::getNeckBehaviourPriority()
{
  return theNeckBehaviourPriority;
}


inline int Neck::getTheIdOfThePlayerToBeSeen()
{
  return theIdOfThePlayerToBeSeen;
}


inline Vector Neck::getThePosToBeSeen()
{
  return thePosToBeSeen;
}


inline double Neck::bodyDir() const
{
  return globalWorldModel->player().bodyDir();
}


inline double Neck::faceDir() const 
{
  return globalWorldModel->player().faceDir();
}

inline int Neck::currentViewWidth() const
{
  return globalWorldModel->player().viewWidth();
}

inline int Neck::currentViewQuality() const
{
  return globalWorldModel->player().viewQuality();
}

inline double Neck::relativeFaceDir()
{
  return globalWorldModel->player().relativeFaceDir();
}

inline double Neck::headAngleToTurnTo( const Vector& point ) const
{
  return (point - globalWorldModel->player().pos()).angle();
}  


#endif
