#include <stdlib.h>
#include <string>
#include <stream.h>
#include <math.h>

#include "Situation.h"
#include "WorldModelCommunicator.h"
#include "WorldModelCore.h"


WorldModelCommunicator::WorldModelCommunicator( ){}


void WorldModelCommunicator::setWorldModelReference( WorldModelCore* aWorldModelCore  )
{
  theCore = aWorldModelCore;
}


void WorldModelCommunicator::writeBitStringForDebug(char *bitstring, int bytes) {
  for (int i=0; i<bytes; ++i) cout << bitstring[i];
  cout <<endl;
}


int WorldModelCommunicator::getByteFromBitstring(const char *bitstring, const int offset) {
  int byte=0;
  for (int i=0; i<8; ++i) if (bitstring[offset+i]=='1') byte+=(int)pow(2,7-i);
  return byte;
}


int WorldModelCommunicator::getCharFromBitstring(const char *bitstring, const int offset) {
  int byte=0;
  for (int i=0; i<5; ++i) if (bitstring[offset+i]=='1') byte+=(int)pow(2,4-i);
  return byte;
}


int WorldModelCommunicator::writeIntToBitstring(int value, const int bits, const int bottom,
                                                const int top, char *bitstring, const int offset) {
  int range=top-bottom;
  int coderange=(int)pow(2,bits)-1;
  if (range<=0) {
    cerr << "ERROR: writeIntToBitstring: negative or zero range!\n";
    return 0;
  }
  if (value<bottom) value=bottom;
  if (value>top) value=top;
  value-=bottom;
  if (coderange<range) {
    cerr << "WARNING: writeIntToBitstring: insufficient bits for encoding int without loss!\n";
    cerr << "         (try to encode "<<range<<" in "<<pow(2,bits)<<" values)\n";
    value=(int)((float)value*(float)coderange/(float)range+.5);
    if (value>coderange) value=coderange;
  }
  for (int i=0; i<bits; ++i) if (value&(int)pow(2,i)) bitstring[offset+i]='1'; else bitstring[offset+i]='0';
  return 1;
}


int WorldModelCommunicator::readIntFromBitstring(const int bits, const int bottom,
                                                 const int top, const char *bitstring, const int offset) {
  int value=0;
  int range=top-bottom;
  int coderange=(int)pow(2,bits)-1;
  if (range<=0) {
    cerr << "ERROR: readIntFromBitstring: negative or zero range!\n";
    return 0;
  }
  for (int i=0; i<bits; ++i) if (bitstring[offset+i]=='1') value+=(int)pow(2,i);
  if (coderange<range) {
    cerr << "WARNING: readIntFromBitstring: insufficient bits for decoding int without loss!\n";
    cerr << "         (try to encode "<<range<<" in "<<pow(2,bits)<<" values)\n";
    value=(int)((float)value*(float)range/(float)coderange+.5);
  }
  return value+bottom;
}


int WorldModelCommunicator::writeDoubleToBitstring(double value, const int bits, const double bottom, const double top, char *bitstring, const int offset) {
  double range=top-bottom;
  int intvalue=0;
  int coderange=(int)pow(2,bits)-1;
  if (range<=0) {
    cerr << "ERROR: writeDoubleToBitstring: negative or zero range!\n";
    return 0;
  }
  if (value<bottom) value=bottom;
  if (value>top) value=top;
  value-=bottom;
  intvalue=(int)(value*(float)coderange/range+.5);
  if (intvalue>coderange) {
    cerr << "WARNING: writeDoubleFromBitstring: range problem!\n";
    intvalue=coderange;
  }
  for (int i=0; i<bits; ++i) if (intvalue&(int)pow(2,i)) bitstring[offset+i]='1'; else bitstring[offset+i]='0';
  return 1;
}


double WorldModelCommunicator::readDoubleFromBitstring(const int bits, const double bottom,
                                                       const double top, const char *bitstring, int offset) {
  double value=0;
  int intvalue=0;
  int coderange=(int)pow(2,bits)-1;
  double range=top-bottom;
  if (range<=0) {
    cerr << "ERROR: readIntFromBitstring: negative or zero range!\n";
    return 0;
  }
  for (int i=0; i<bits; ++i) if (bitstring[offset+i]=='1') intvalue+=(int)pow(2,i);
  value=(float)intvalue*range/coderange;

  return value+bottom;
}


int WorldModelCommunicator::generateSayMessage(char *message, const char *bitstring, const int bits) {
  static char *convert="0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
  int messagepointer=0;
  for (int i=0; i<bits; i+=5) message[messagepointer++]=convert[getCharFromBitstring(bitstring,i)];
  message[messagepointer]=0;
  return 1;
}


int WorldModelCommunicator::generateBitstring(char *bitstring, const char *message) {
  static char *convert="0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
  int messagepointer=0;
  int offset=0;
  while (message[messagepointer] && messagepointer<512) {
    for (int i=0; i<65; ++i)  {
      if (i>64) {
        cerr <<"WARNING: generateBitString: illegal character in input!\n";
        for (int j=0; j<5; ++j) bitstring[offset++]='0';
      }
      if (message[messagepointer]==convert[i]) {
        for (int j=0; j<5; ++j) {
          if (i&(int)pow(2,4-j)) bitstring[offset]='1'; else bitstring[offset]='0';
          offset++;
        }
      }
    }
    messagepointer++;
  }

  return offset;
}


int WorldModelCommunicator::calculateChecksum(char *bitstring, int bits) {
  // berechnet die 8-Bit-Prfsumme der ersten bits/8 bits (dh nur ganze Bytes)
  int checksum=0;
  for (int i=0; i<bits-7; i+=8) checksum+=getByteFromBitstring(bitstring,i);
  return checksum%256;
}


Ath98String WorldModelCommunicator::saySituation(const Situation& situation) {

  const int say_msg_size=512;

  static char message[say_msg_size+1];
  static char bitstring[say_msg_size*8+10];


  if (situation.simStep()==UNDEF_INT) {
    cerr << "WARNING: saySituation called with undef situation!\n";
    return Ath98String();
  }

  int offset=0;

  writeIntToBitstring('Z',8,0,255,bitstring,offset);    // Zorro was here
  offset+=8;  
  writeIntToBitstring(theCore->side(),2,200,203,bitstring,offset);    // Seite mit codieren
  offset+=8;
  writeIntToBitstring(situation.simStep(),16,0,65535,bitstring,offset);
  offset+=16;
  writeIntToBitstring(theCore->id(),6,MIN_PLAYER_ID,MAX_PLAYER_ID,bitstring,offset);
  offset+=6;
  writeIntToBitstring(situation.playerCount(),7,0,31,bitstring,offset);
  offset+=7;
  for (int i=0; i<situation.playerCount(); i++) {
    writeDoubleToBitstring(situation.playerWithIndex(i).pos().x(),14,-55.0,+55.0,bitstring,offset);
    offset+=14;
    writeDoubleToBitstring(situation.playerWithIndex(i).pos().y(),13,-36.0,+36.0,bitstring,offset);
    offset+=13;
    writeDoubleToBitstring(situation.playerWithIndex(i).posConfidence().reliability(),6,0.0,1.0,bitstring,offset);
    offset+=6;
    writeDoubleToBitstring(situation.playerWithIndex(i).speed().x(),9,0.0,+2.0,bitstring,offset);
    offset+=9;
    writeDoubleToBitstring(situation.playerWithIndex(i).speed().y(),9,0.0,+2.0,bitstring,offset);
    offset+=9;
    writeDoubleToBitstring(situation.playerWithIndex(i).speedConfidence().reliability(),6,0.0,1.0,bitstring,offset);
    offset+=6;
    writeIntToBitstring(situation.playerWithIndex(i).id(),6,MIN_PLAYER_ID,MAX_PLAYER_ID,bitstring,offset);
    offset+=6;
    writeDoubleToBitstring(situation.playerWithIndex(i).playerIdConfidence().reliability(),6,0.0,1.0,bitstring,offset);
    offset+=6;
    writeDoubleToBitstring(situation.playerWithIndex(i).teamIdConfidence().reliability(),6,0.0,1.0,bitstring,offset);
    offset+=6;
  }
  writeDoubleToBitstring(situation.ball().pos().x(),14,-55.0,+55.0,bitstring,offset);
  offset+=14;
  writeDoubleToBitstring(situation.ball().pos().y(),13,-36.0,+36.0,bitstring,offset);
  offset+=13;
  writeDoubleToBitstring(situation.ball().posConfidence().reliability(),6,0.0,1.0,bitstring,offset);
  offset+=6;
  writeDoubleToBitstring(situation.ball().speed().x(),9,0.0,+2.7,bitstring,offset);
  offset+=9;
  writeDoubleToBitstring(situation.ball().speed().y(),9,0.0,+2.7,bitstring,offset);
  offset+=9;
  writeDoubleToBitstring(situation.ball().speedConfidence().reliability(),6,0.0,1.0,bitstring,offset);
  offset+=6;

  int checksum=calculateChecksum(bitstring,offset);
  writeIntToBitstring(checksum,8,0,255,bitstring,offset);
  offset+=8;

  generateSayMessage(message,bitstring,offset);

  // Message abschicken

  return Ath98String(message);
}


int WorldModelCommunicator::understandYourSituation(Situation& situation, const char* message) {

  const int say_msg_size=512;

  static char bitstring[say_msg_size*8+10];
  int length=generateBitstring(bitstring, message);

  int offset=0;

  if (readIntFromBitstring(8,0,255,bitstring,offset) != 'Z') {  // Zorro was not here :-(
//    cerr<< "WARNING: understandYourSituation called with alien message!\n";
    return 0;
  }
  offset+=8;
  if( readIntFromBitstring(2,200,203,bitstring,offset) != theCore->side() ){  // Zorro spielt auf der falschen Seite
//		cerr<< "WARNING: understandYourSituation called with alien message! Zorro plays on wrong side!\n";
    return 0;
  }    
  offset+=8;
  situation.setSimStep(readIntFromBitstring(16,0,65535,bitstring,offset));
  offset+=16;
  int callerId=readIntFromBitstring(6,MIN_PLAYER_ID,MAX_PLAYER_ID,bitstring,offset);
  offset+=6;
  int numberOfPlayers=readIntFromBitstring(7,0,31,bitstring,offset);
  offset+=7;
  
  for (int i=0; i<numberOfPlayers; i++) {
    ModelledPlayer player;
    // Position
    double x=readDoubleFromBitstring(14,-55.0,+55.0,bitstring,offset);
    offset+=14;
    double y=readDoubleFromBitstring(13,-36.0,+36.0,bitstring,offset);
    offset+=13;
    player.setPos(Vector(x,y));
    
    // PositionConfidence
    double reliability=readDoubleFromBitstring(6,0.0,1.0,bitstring,offset);
    offset+=6;
    player.setPosConfidence(Confidence(situation.simStep(),2.0,reliability));

    // Speed 
    double xSpeed=readDoubleFromBitstring(9,0.0,+2.0,bitstring,offset);
    offset+=9;
    double ySpeed=readDoubleFromBitstring(9,0.0,+2.0,bitstring,offset);
    offset+=9;
    player.setSpeed(Vector(xSpeed,ySpeed));
    
    // SpeedConfidence
    reliability=readDoubleFromBitstring(6,0.0,1.0,bitstring,offset);
    offset+=6;
    player.setSpeedConfidence(Confidence(situation.simStep(),0.7,reliability));

    // Id
    player.setId(readIntFromBitstring(6,MIN_PLAYER_ID,MAX_PLAYER_ID,bitstring,offset));
    offset+=6;

    // idConfidence
    reliability=readDoubleFromBitstring(6,0.0,1.0,bitstring,offset);
    offset+=6;
    player.setPlayerIdConfidence(Confidence(situation.simStep(),1.0,reliability));

    // teamIdConfidence
    reliability=readDoubleFromBitstring(6,0.0,1.0,bitstring,offset);
    offset+=6;
    player.setTeamIdConfidence(Confidence(situation.simStep(),1.0,reliability));

    if( (   (player.id() >= MIN_TEAMMATE_ID)
	 && (player.id() <= MAX_EXISTING_TEAMMATE_ID) )
	||
	(   (player.id() >= MIN_OPPONENT_ID)
	 && (player.id() <= MAX_OPPONENT_ID) )
 	) {
      if(player.id() != theCore->id()) {
         situation.setPlayer(player);
      }
    }
  }

  ModelledBall ball;
  ball.setId(BALL_ID);
  // Position
  double x=readDoubleFromBitstring(14,-55.0,+55.0,bitstring,offset);
  offset+=14;
  double y=readDoubleFromBitstring(13,-36.0,+36.0,bitstring,offset);
  offset+=13;
  ball.setPos(Vector(x,y));

  // PositionConfidence
  double reliability=readDoubleFromBitstring(6,0.0,1.0,bitstring,offset);
  offset+=6;
  ball.setPosConfidence(Confidence(situation.simStep(),2.0,reliability));

  // Speed 
  double xSpeed=readDoubleFromBitstring(9,0.0,+2.7,bitstring,offset);
  offset+=9;
  double ySpeed=readDoubleFromBitstring(9,0.0,+2.7,bitstring,offset);
  offset+=9;
  ball.setSpeed(Vector(xSpeed,ySpeed));

  // SpeedConfidence
  reliability=readDoubleFromBitstring(6,0.0,1.0,bitstring,offset);
  offset+=6;
  if(   ( theCore->player().isInFieldOfView( ball ) == YES )
     && ( situation.playerWithId( callerId ).distTo( ball.pos() ) > theCore->player().distTo( ball.pos() ) ) ){
    situation.setBall( theCore->ball() );
  }
  else{
    ball.setSpeedConfidence(Confidence(situation.simStep(),0.7,reliability));
    situation.setBall(ball);
  }

  int checksum=calculateChecksum(bitstring,offset);
  if (checksum!=readIntFromBitstring(8,0,255,bitstring,offset)) {
    cerr << "WARNING: understandYourSituation called with wrong checksum!\n";
    return 0;
  }
  offset+=8;

  // Situation in die Gegenwart simulieren

  int stepDifference=theCore->simStep()-situation.simStep();
  if (stepDifference>3) {
    cerr << "WARNING: understandYourSituation called with prehistoric situation!\n";
    return 0;
  }
  if (stepDifference<0) {
    cerr << "WARNING: understandYourSituation called with futuristic situation!\n";
    return 0;
  }

  if( stepDifference > 0 ){
    situation.simulateSteps(stepDifference);
  }

  return 1;
}

