#ifndef _WORLDMODELINTEGRATOR_H_
#define _WORLDMODELINTEGRATOR_H_

class ModelledBall;
class ModelledPlayer;
class Message;
class Action;
class Vector;
class WorldModelCore;
class BodyInfos;
class Situation;
class Sensors;


class WorldModelIntegrator 
{
  friend WorldModelCore;
 
 private:
  
  WorldModelCore* theCore;

  Sensors* theSensors;

  WorldModelIntegrator( );
  
  void setWorldModelReference( WorldModelCore* aWorldModelCore  );  
  
  void setSensorsReference( Sensors* theSensors  ); 

   /** @name Einarbeitung neuer Informationen */
  //@{
    /** Integriert die Sensorinformationen in das Weltmodell.
        Bei einer neuen BodyInformation wird die Methode
        \Ref{processBodyInfo} ausgefuehrt.\\
        Bei einer neuen Schiedsrichterentscheidung wird die Methode
        \Ref{processRefereeInfo} ausgefuehrt.\\
        Bei einer neuen Hoerinformation, die nicht vom Schiedsrichter
        ist, wird die Methode \Ref{processHearInfo} ausgefuehrt und
        die dadurch erzeugte Situation mit der aktuellen Situation
        mittels der Methode \Ref{mergeSituations} fusioniert. \\
        Bei einer neuen Sichtinformation wird die Methode
        \Ref{processSeeInfo} ausgefuehrt und die dadurch erzeugte
        Situation mit der aktuellen mittels der Methode 
        \Ref{mergeSituations}fusioniert. Danach wird die
        fusionierte Situation als neue aktuelle Situation angenommen.\\
        Es koennen mehrere dieser Schritte hintereinander durchgefuehrt
        werden, falls mehrere neue Informationen vorliegen.\\
        Am Ende wird mit der Methode \Ref{improveConsistency} ein
        Konsistenzcheck auf dem Weltmodell und der aktuellen Situation
        durchgefuehrt und die Konsistenz wieder hergestellt, falls
        dies noetig sein sollte. \\
        @see processBodyInfo
        @see processRefereeInfo
        @see processHearInfo
        @see processSeeInfo
        @see mergeSituations
        @see improveConsistency */
    void integrateSensorInfo();

    /** Integriert die korrekten Sensorinformationen (Fullstate) in einen 
        separaten Bereich des Weltmodells. */        
    void integrateCorrectSensorInfo();

    /** Registriert, da\3 eine Aktion abgesendet wurde.
        Die abgesendete Aktion wird gepuffert. Eine Simulation wird
        erst durchgefuehrt, wenn eine Bestaetigung der Ausfuehrung
        durch den SoccerServer vorliegt (durch BodyInformation).
        @param anAction Die abgeschickte Aktion */
    void registerExecution( const Action& anAction );

 /** Integriert die Sichtinformationen ins Weltmodell.
   */
  void processSeeInfo( const BodyInfos& bodyInfosForSee ); 
  
  /** Speichert eine Situation als die (neue) gegenwaertige ab. 
      Die bisherige gegenwaertige Situation ist nach dem Aufruf mit 
      situation( -1 )  erreichbar. 
      @param aSituation Die zu speichernde Situation */  
  void addCurrentSituation( const Situation& aSituation ); 
  
  /** Entfernt die aktuelle Situation aus dem Weltmodell. 
      Dies kann sinnvoll sein, wenn man aus Versehen zu weit in die Zukunft
      simuliert hat und den aktuellen Stand wieder herstellen m\"ochte. */
  void removeCurrentSituation();
  
  /** Berechnet die relative Position des Objekts. 
      Benoetigt werden die gesehene Position des Objektes und der eigene 
      Blickwinkel. 
      @param seenPos Vektor, der von der eigenen Position zum Objekt zeigt.
      Er ist also zu interpretieren in einem Koordinatensystem, dessen
      y-Achse in die Blickrichtung zeigt und in dessen Ursprung die eigene
      Position liegt.
      @param Der eigene Blickwinkel 
      @return die relative Position des Objektes */
  Vector relativePos( const Vector& seenPos, double faceDir );
  
  /** Setzt den simStep, an dem das letzte Catch-Kommando an den Server
      geschickt wurde. (Mu\3 von au\3en setzbar sein, damit der Goalie,
      nachdem er den Ball freigegeben hat, nicht gleich wieder catcht. */
  void setSimStepOfLastCatch( int newSimStep );
  
  /** Setzt den simStep, an dem das letzte erfolgreiche Catchkommando
      ausgef/"uhrt wurde. */
  void setSimStepWhenCatchedBallIsReleased( int newSimStep );
  
  
  /** Setzt den simStep, an dem das letzte Say-Kommando an den Server
      geschickt wurde. */
  void setSimStepOfLastSay( int newSimStep );
  
  /** Setzt den simStep, an dem die letzte Schiedsrichterentscheidung
      eingetroffen ist. */
  void setSimStepOfLastRefereeDecision( int newSimStep );
  
  /** Setzt den simStep, an dem die letzte Freekick-Entscheidung
      eingetroffen ist. */
  void setSimStepOfLastFreekick( int newSimStep );

  /** Setzt den simStep, an dem die letzte See-Info
      eingetroffen ist. */
  void setSimStepOfLastSee( int newSimStep );
  
  /** Integriert eine neue BodyInformation. Wenn die BodyInformationen
      einen neuen Simulationstakt enthalten werden alle an den SoccerServer
      gesendeten und mit dieser BodyInformation bestaetigten Kommandos
      im Agenten simuliert und eine neue Situation mit den aktuellen
      Simulationstakt angelegt. Alle nicht bestaetigten Actions werden
      als nicht vom SoccerServer empfangen gewertet. Zum Simulieren
      der Actions wird die Methode \Ref{updateWithAction} aufgerufen. \\
      @see updateWithAction */
  void processBodyInfo( const BodyInfos& bodyInfos );
  
  /** Korrigiert die Situationen des Weltmodells, so dass neue Informationen
      aus den Sensoren korrekt integriert werden koennen. Es werden, wenn
      noetig bereits ebgesendete und gequeute Befehle simuliert bis der
      im Parameter angegebene SimStep erreicht ist.
  */
  void synchronizeSituations( int simStep );
  
  /** Ruft fuer alle neu eingegangenen Nachrichten vom Schiedsrichter die
      Methode \Ref{processRefereeMessage} auf.
      @see processRefereeMessage */
  void processRefereeInfo();
  
  /** Ruft fuer alle neu eingegangenen Nachrichten von Mitspielern die
      Methode \Ref{processOurTeamMessage} auf und fusioniert die dadurch
      erlangten Informationen mit der aktuellen Situation des Weltmodells.
      Wurden Rufe mehrerer Spieler empfangen werden deren Daten vorher mit
      der Methode \Ref{mergeSituations} fusioniert.
      @see processOurTeamMessage */
  void processHearInfo();
  
  /** Integriert die Sichtinformationen ins Weltmodell.
   */
  // void processSeeInfo( const BodyInfos& bodyInfosForSee );
  
  /** Stellt die Konsistenz des Weltmodells wieder her. In Abhaengigkeit
      vom aktuellen Spielzustand werden einige Objekte der aktuellen
      Situation umgesetzt.
  */
  void improveConsistency();
  
  /** Integriert die Folgen der Ausf\"uhrung einer Aktion.
      Ist die Aktion eine Hauptaktion, wird eine neue Situation unter der
      Annahme erzeugt, dass der Agent in der aktuellen Situation die
      angegebene Aktion ausgefuehrt hat. Die neue Situation wird zur
      aktuellen Situation. Ist die Aktion eine Nebenaktion, wird keine
      neue Situation angelegt sondern die aktuelle veraendert. 
      @param Die Action, die integriert werden soll. */
  void updateWithAction( const Action& anAction );  
  
  /** Verarbeitet Schiedsrichterentscheidungen. 
      Dabei werden Spielzustand und Initiative neu gesetzt.
      @param message Die Nachricht des Schiedsrichters */  
  void processRefereeMessage( const Message& message );
  
  /** F\"ugt Informationen aus der anderen Situation in die 
      vertrauensw\"urdigere Situation ein.
      Die \Ref{Reliablity} Objekte bilden die Grundlage f\"ur das Fusionieren
      der einzelnen Objekte. Verglichen werden nur Objekte die gleich sind.
      Es werden jeweils die aktuelleren Daten dieser Objekte \"ubernommen,
      bei Gleichheit siegt das Datum, da\3 der vertrauensw\"urdigeren
      Situation entspricht. Daten der anderen Situation, die keine
      Entsprechung in der vertauensw\"urdigen Situation haben werden nicht
      \"ubernommen.
      @param trustedSituation vertrauensw\"urdige Situation dessen Daten als
      Grundlage angesehen werden.
      @param anotherSituation andere Situation, dessen Daten einverleibt
      werden sollen. 
      @see mergeBalls
  */
    Situation mergeSituations( const Situation& trustedSituation,
                          const Situation& anotherSituation, bool mergeWithSeenSituation = NO );
    /** Liefert einen \Ref{ModelledPlayer}, der die Daten
        \"uber den Spieler selbst enth\"alt.
        Diese Daten stammen entweder aus den gesehenen Informationen oder,
        wenn diese Informationen nicht ausreichen, aus den alten Situation.
        @return ModelledPlayer Spieler mit Daten \"uber den Spieler selbst */
    const ModelledPlayer modelOfSelf( const BodyInfos& bodyInfos ) const;
    
    /** F\"ugt alle gesehenen Objekte in die Situation ein.
        Diese Daten entstammen nur den gesehenen Informationen.
        @param newSituation Situation in die Daten eingef\"ugt werden
        sollen. */
    void addSeenObjects( Situation& newSituation );
    
    /** Aktualisiert die Werte des uebergebenen Spielers mit den 
        uebergebenen BodyInfos wenn es passende BodyInfos sind
        (gleicher simstep wie weltmodell) */
    void updateOwnPlayerWithBodyInfos( ModelledPlayer& playerSelf,
                                       const BodyInfos& bodyInfos ) const;
    
    /** Dieser Algorithmus wird gerufen, wenn ein Ball gesehen wurde */
    ModelledBall trackBall(const ModelledBall& absoluteSeenBall, const ModelledPlayer& myself);
    
};

#endif
